@@ -33,7 +33,8 @@ use 5.008;  # We're using PerlIO layers
 use strict;
 use open ':utf8';  # changelogs are written with UTF-8 encoding
 use filetest 'access';  # use access rather than stat for -w
-use Encode 'decode_utf8';  # for checking whether user names are valid
+# for checking whether user names are valid and making format() behave
+use Encode qw/decode_utf8 encode_utf8/;
 use Getopt::Long;
 use File::Copy;
 use File::Basename;
@@ -60,11 +61,11 @@ sub have_lpdc {
     };
 
     if ($@) {
-        if ($@ =~ m%^Can\'t locate Parse/DebControl%) {
-            $lpdc_broken="the libparse-debcontrol-perl package is not installed";
-        } else {
-            $lpdc_broken="couldn't load Parse::DebControl: $@";
-        }
+	if ($@ =~ m%^Can\'t locate Parse/DebControl%) {
+	    $lpdc_broken="the libparse-debcontrol-perl package is not installed";
+	} else {
+	    $lpdc_broken="couldn't load Parse::DebControl: $@";
+	}
     }
     else { $lpdc_broken=''; }
     return $lpdc_broken ? 0 : 1;
@@ -87,6 +88,16 @@ Options:
          Update the changelog timestamp. If the distribution is set to
          "UNRELEASED", change it to unstable (or another distribution as
          specified by --distribution).
+  --force-save-on-release
+         When --release is used and an editor opened to allow inspection
+         of the changelog, require the user to save the changelog their
+         editor opened.  Otherwise, the original changelog will not be
+         modified. (default)
+  --no-force-save-on-release
+         Do not do so. Note that a dummy changelog entry made be supplied
+         in order to achieve the same effect - e.g. $progname --release ""
+         The entry will not be added to the changelog but its presence will
+         suppress the editor
   --create
          Create a new changelog (default) or NEWS file (with --news) and
          open for editing
@@ -104,6 +115,8 @@ Options:
          Increment the Debian release number for a non-maintainer upload
   --bin-nmu
          Increment the Debian release number for a binary non-maintainer upload
+  -R, --rebuild
+         Increment the Debian release number for a no-change rebuild
   -q, --qa
          Increment the Debian release number for a Debian QA Team upload
   -s, --security
@@ -112,7 +125,7 @@ Options:
          Increment the Debian release number without any distribution name
   --bpo
          Increment the Debian release number for a Backports.org upload
-	 to "etch-backports"
+	 to "lenny-backports"
   -l, --local <suffix>
          Add a suffix to the Debian version number for a local build
   -b, --force-bad-version
@@ -163,7 +176,7 @@ Options:
          What constitutes a matching directory name; REGEX is
          a Perl regular expression; the string \`PACKAGE\' will
          be replaced by the package name; see manpage for details
-         (default: 'PACKAGE(-.*)?')
+         (default: 'PACKAGE(-.+)?')
   --no-conf, --noconf
          Don\'t read devscripts config files; must be the first option given
   --release-heuristic log|changelog
@@ -173,7 +186,7 @@ Options:
          Display this help message and exit
   --version
          Display version information
-  At most one of -a, -i, -e, -r, -v, -d, -n, --bin-nmu, -q, --qa, -s, --bpo, -l
+  At most one of -a, -i, -e, -r, -v, -d, -n, --bin-nmu, -R, -q, --qa, -s, --bpo, -l
   (or their long equivalents) may be used.
   With no options, one of -i or -a is chosen by looking for a .upload
   file in the parent directory and checking its contents.
@@ -196,7 +209,7 @@ EOF
 
 # Start by setting default values
 my $check_dirname_level = 1;
-my $check_dirname_regex = 'PACKAGE(-.*)?';
+my $check_dirname_regex = 'PACKAGE(-.+)?';
 my $opt_p = 0;
 my $opt_query = 1;
 my $opt_release_heuristic = 'log';
@@ -206,6 +219,7 @@ my $opt_tz = undef;
 my $opt_t = '';
 my $opt_allow_lower = '';
 my $opt_auto_nmu = 'yes';
+my $opt_force_save_on_release = 1;
 
 # Next, read configuration files and then command line
 # The next stuff is boilerplate
@@ -219,7 +233,7 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 		       'DEBCHANGE_PRESERVE' => 'no',
 		       'DEBCHANGE_QUERY_BTS' => 'yes',
 		       'DEVSCRIPTS_CHECK_DIRNAME_LEVEL' => 1,
-		       'DEVSCRIPTS_CHECK_DIRNAME_REGEX' => 'PACKAGE(-.*)?',
+		       'DEVSCRIPTS_CHECK_DIRNAME_REGEX' => 'PACKAGE(-.+)?',
 		       'DEBCHANGE_RELEASE_HEURISTIC' => 'log',
 		       'DEBCHANGE_MULTIMAINT' => 'yes',
 		       'DEBCHANGE_TZ' => $ENV{TZ}, # undef if TZ unset
@@ -227,10 +241,11 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 		       'DEBCHANGE_MAINTTRAILER' => '',
 		       'DEBCHANGE_LOWER_VERSION_PATTERN' => '',
 		       'DEBCHANGE_AUTO_NMU' => 'yes',
+		       'DEBCHANGE_FORCE_SAVE_ON_RELEASE' => 'yes',
 		       );
     $config_vars{'DEBCHANGE_TZ'} ||= '';
     my %config_default = %config_vars;
-    
+
     my $shell_cmd;
     # Set defaults
     foreach my $var (keys %config_vars) {
@@ -258,6 +273,8 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 	or $config_vars{'DEBCHANGE_MULTIMAINT_MERGE'}='no';
     $config_vars{'DEBCHANGE_AUTO_NMU'} =~ /^(yes|no)$/
 	or $config_vars{'DEBCHANGE_AUTO_NMU'}='yes';
+    $config_vars{'DEBCHANGE_FORCE_SAVE_ON_RELEASE'} =~ /^(yes|no)$/
+	or $config_vars{'DEBCHANGE_FORCE_SAVE_ON_RELEASE'}='yes';
 
     foreach my $var (sort keys %config_vars) {
 	if ($config_vars{$var} ne $config_default{$var}) {
@@ -279,13 +296,15 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 	if $config_vars{'DEBCHANGE_MAINTTRAILER'};
     $opt_allow_lower = $config_vars{'DEBCHANGE_LOWER_VERSION_PATTERN'};
     $opt_auto_nmu = $config_vars{'DEBCHANGE_AUTO_NMU'};
+    $opt_force_save_on_release =
+	$config_vars{'DEBCHANGE_FORCE_SAVE_ON_RELEASE'} eq 'yes' ? 1 : 0;
 }
 
 # We use bundling so that the short option behaviour is the same as
 # with older debchange versions.
 my ($opt_help, $opt_version);
 my ($opt_i, $opt_a, $opt_e, $opt_r, $opt_v, $opt_b, $opt_d, $opt_D, $opt_u, $opt_force_dist);
-my ($opt_n, $opt_bn, $opt_qa, $opt_s, $opt_U, $opt_bpo, $opt_l, $opt_c, $opt_m, $opt_create, $opt_package, @closes);
+my ($opt_n, $opt_bn, $opt_R, $opt_qa, $opt_s, $opt_U, $opt_bpo, $opt_l, $opt_c, $opt_m, $opt_create, $opt_package, @closes);
 my ($opt_news);
 my ($opt_ignore, $opt_level, $opt_regex, $opt_noconf, $opt_empty);
 my ($opt_distributor);
@@ -311,6 +330,7 @@ GetOptions("help|h" => \$opt_help,
 	   "u|urgency=s" => \$opt_u,
 	   "n|nmu" => \$opt_n,
 	   "bin-nmu" => \$opt_bn,
+	   "R|rebuild" => \$opt_R,
 	   "q|qa" => \$opt_qa,
 	   "s|security" => \$opt_s,
 	   "U|upstream" => \$opt_U,
@@ -332,6 +352,7 @@ GetOptions("help|h" => \$opt_help,
 	   "release-heuristic=s" => \$opt_release_heuristic,
 	   "empty" => \$opt_empty,
 	   "auto-nmu!" => \$opt_auto_nmu,
+	   "force-save-on-release!" => \$opt_force_save_on_release,
 	   )
     or die "Usage: $progname [options] [changelog entry]\nRun $progname --help for more details\n";
 
@@ -367,8 +388,8 @@ if (defined $opt_level) {
 if (defined $opt_regex) { $check_dirname_regex = $opt_regex; }
 
 # Only allow at most one non-help option
-fatal "Only one of -a, -i, -e, -r, -v, -d, -n/--nmu, --bin-nmu, -q/--qa, -s/--security, --bpo, -l/--local is allowed;\ntry $progname --help for more help"
-    if ($opt_i?1:0) + ($opt_a?1:0) + ($opt_e?1:0) + ($opt_r?1:0) + ($opt_v?1:0) + ($opt_d?1:0) + ($opt_n?1:0) + ($opt_bn?1:0) + ($opt_qa?1:0) + ($opt_s?1:0) + ($opt_bpo?1:0) + ($opt_l?1:0) > 1;
+fatal "Only one of -a, -i, -e, -r, -v, -d, -n/--nmu, --bin-nmu, -R/--rebuild, -q/--qa, -s/--security, --bpo, -l/--local is allowed;\ntry $progname --help for more help"
+    if ($opt_i?1:0) + ($opt_a?1:0) + ($opt_e?1:0) + ($opt_r?1:0) + ($opt_v?1:0) + ($opt_d?1:0) + ($opt_n?1:0) + ($opt_bn?1:0) + ($opt_R?1:0) + ($opt_qa?1:0) + ($opt_s?1:0) + ($opt_bpo?1:0) + ($opt_l?1:0) > 1;
 
 if ($opt_s) {
     $opt_u = "high";
@@ -395,13 +416,13 @@ if ($distributor eq 'Ubuntu' and ($opt_n or $opt_bn or $opt_qa or $opt_bpo)) {
 # Check the distro name given.
 if (defined $opt_D) {
     if ($distributor eq 'Debian') {
-	unless ($opt_D =~ /^(unstable|((old)?stable|testing)(-security)?|experimental|UNRELEASED|(sarge|etch)(-volatile|-backports)?|((oldstable|testing)-)?proposed-updates)$/) {
-	    warn "$progname warning: Recognised distributions are: unstable, testing, stable,\noldstable, experimental, UNRELEASED, {sarge,etch}-{volatile,backports},\n{oldstable-,testing-,}proposed-updates and {testing,stable,oldstable}-security.\nUsing your request anyway.\n";
+	unless ($opt_D =~ /^(unstable|((old)?stable|testing)(-security)?|experimental|UNRELEASED|(sarge|etch|lenny)(-volatile|-backports)|((oldstable|testing)-)?proposed-updates)$/) {
+	    warn "$progname warning: Recognised distributions are: unstable, testing, stable,\noldstable, experimental, UNRELEASED, {sarge,etch,lenny}-{volatile,backports},\n{oldstable-,testing-,}proposed-updates and {testing,stable,oldstable}-security.\nUsing your request anyway.\n";
 	    $warnings++ if not $opt_force_dist;
 	}
     } elsif ($distributor eq 'Ubuntu') {
-	unless ($opt_D =~ /^((warty|hoary|breezy|dapper|edgy|feisty|gutsy|hardy|intrepid|jaunty)(-updates|-security|-proposed)?|UNRELEASED)$/) {
-	    warn "$progname warning: Recognised distributions are:\n{warty,hoary,breezy,dapper,edgy,feisty,gutsy,hardy,intrepid,jaunty}{,-updates,-security,-proposed} and UNRELEASED.\nUsing your request anyway.\n";
+	unless ($opt_D =~ /^((dapper|hardy|intrepid|jaunty|karmic|lucid|maverick)(-updates|-security|-proposed|-backports)?|UNRELEASED)$/) {
+	    warn "$progname warning: Recognised distributions are:\n{dapper,hardy,intrepid,jaunty,karmic,lucid,maverick}{,-updates,-security,-proposed,-backports} and UNRELEASED.\nUsing your request anyway.\n";
 	    $warnings++ if not $opt_force_dist;
 	}
     } else {
@@ -411,9 +432,10 @@ if (defined $opt_D) {
 
 fatal "--closes should not be used with --news; put bug numbers in the changelog not the NEWS file"
     if $opt_news && @closes;
-    
-fatal "--package can only be used with --create"
-    if $opt_package && ! $opt_create;
+
+# hm, this can probably be used with more than just -i.
+fatal "--package can only be used with --create, --increment and --newversion"
+    if $opt_package && ! ($opt_create || $opt_i || $opt_v);
 
 my $changelog_path = $opt_c || $ENV{'CHANGELOG'} || 'debian/changelog';
 my $real_changelog_path = $changelog_path;
@@ -428,8 +450,8 @@ fatal "--package cannot be used when creating a NEWS file"
 
 if ($opt_create) {
     if ($opt_a || $opt_i || $opt_e || $opt_r || $opt_b || $opt_n || $opt_bn ||
-	    $opt_qa || $opt_s || $opt_bpo || $opt_l || $opt_allow_lower) {
-	warn "$progname warning: ignoring -a/-i/-e/-r/-b/--allow-lower-version/-n/--bin-nmu/-q/--qa/-s/--bpo/-l options with --create\n";
+	    $opt_R || $opt_qa || $opt_s || $opt_bpo || $opt_l || $opt_allow_lower) {
+	warn "$progname warning: ignoring -a/-i/-e/-r/-b/--allow-lower-version/-n/--bin-nmu/-R/-q/--qa/-s/--bpo/-l options with --create\n";
 	$warnings++;
     }
     if ($opt_package && $opt_d) {
@@ -496,7 +518,11 @@ my $MAINTAINER = 'MAINTAINER';
 my $EMAIL = 'EMAIL';
 my $DISTRIBUTION = 'UNRELEASED';
 my $bpo_dist = '';
+my %bpo_dists = ( '40', 'etch', '50', 'lenny' );
+my $latest_bpo_dist = '50';
 my $CHANGES = '';
+# Changelog urgency, possibly propogated to NEWS files
+my $CL_URGENCY = '';
 
 if (! $opt_create || ($opt_create && $opt_news)) {
     if (! $opt_create) {
@@ -539,8 +565,8 @@ if (! $opt_create || ($opt_create && $opt_news)) {
     fatal "No distribution in changelog!"
 	unless exists $changelog{'Distribution'};
     if ($distributor eq 'Ubuntu') {
-	# In Ubuntu uploads should go to intrepid
-	$DISTRIBUTION = 'intrepid';
+	# In Ubuntu uploads should go to lucid
+	$DISTRIBUTION = 'lucid';
     } else {
 	$DISTRIBUTION=$changelog{'Distribution'};
     }
@@ -549,12 +575,22 @@ if (! $opt_create || ($opt_create && $opt_news)) {
 
     # Find the current package version
     if ($opt_news) {
+	my $found_version = 0;
+	my $found_urgency = 0;
 	open PARSED, qq[dpkg-parsechangelog -l"$real_changelog_path" | ]
 	    or fatal "Cannot execute dpkg-parsechangelog: $!";
 	while (<PARSED>) {
 	    chomp;
 	    if (m%^Version:\s+(\S+)$%) {
 		$VERSION = $1;
+		$VERSION =~ s/~$//;
+		$found_version = 1;
+		last if $found_urgency;
+	    } elsif (m%^Urgency:\s+(\S+)(\s|$)%) {
+		$CL_URGENCY = $1;
+		$found_urgency = 1;
+		last if $found_version;
+	    } elsif (m%^$%) {
 		last;
 	    }
 	}
@@ -600,21 +636,22 @@ EOF
 	    # don't know anything
 	}
     }
-    if ($opt_package) {
-	if ($opt_package =~ m/^[a-z0-9][a-z0-9+\-\.]+$/) {
-	    $PACKAGE=$opt_package;
-	} else {
-	    warn "$progname warning: illegal package name used with --package: $opt_package\n";
-	    $warnings++;
-	}
-    }
     if ($opt_v) {
 	$VERSION=$opt_v;
     }
     if ($opt_D) {
 	$DISTRIBUTION=$opt_D;
     }
-}    
+}
+
+if ($opt_package) {
+    if ($opt_package =~ m/^[a-z0-9][a-z0-9+\-\.]+$/) {
+	$PACKAGE=$opt_package;
+    } else {
+	warn "$progname warning: illegal package name used with --package: $opt_package\n";
+	$warnings++;
+    }
+}
 
 # Clean up after old versions of debchange
 if (-f "debian/RELEASED") {
@@ -658,7 +695,7 @@ if (! $opt_m) {
 	$MAINTAINER = $env{'NAME'};
     } else {
 	my @pw = getpwuid $<;
-	if (defined($pw[6])) {
+	if ($pw[6]) {
 	    if (my $pw = decode_utf8($pw[6])) {
 		$pw =~ s/,.*//;
 		$MAINTAINER = $pw;
@@ -701,9 +738,12 @@ if (! $opt_m) {
 #####
 
 if ($opt_auto_nmu eq 'yes' and ! $opt_v and ! $opt_l and ! $opt_s and 
-    ! $opt_qa and ! $opt_bpo and ! $opt_bn and ! $opt_n and ! $opt_c and
-    ! $opt_create and ! $opt_a_passed and ! $opt_r and
-    $distributor ne 'Ubuntu') {
+    ! $opt_qa and ! $opt_bpo and ! $opt_bn and ! $opt_R and ! $opt_n and ! $opt_c and
+    ! (exists $ENV{'CHANGELOG'} and length $ENV{'CHANGELOG'}) and
+    ! $opt_create and ! $opt_a_passed and ! $opt_r and ! $opt_e and
+    $distributor ne 'Ubuntu' and
+    ! ($opt_release_heuristic eq 'changelog' and
+	$changelog{'Distribution'} eq 'UNRELEASED' and ! $opt_i_passed)) {
 
     if (-f 'debian/control') {
 	if (have_lpdc()) {
@@ -745,23 +785,20 @@ if (@closes and $opt_query) { # and we have to query the BTS
 	my $bugs = Devscripts::Debbugs::select( "src:" . $PACKAGE );
 	my $statuses = Devscripts::Debbugs::status(
 	    map {[bug => $_, indicatesource => 1]} @{$bugs} );
-	if (not defined $bugs) {
-	    warn "$progname warning: Debbugs::status failed, so cannot query the bug-tracking system\n";
-	    $opt_query=0;
-	    $warnings++;
-	    # This will now go and execute the "if (@closes and ! $opt_query)" code
+	if ($statuses eq "") {
+	    warn "$progname: No bugs found for package $PACKAGE\n";
 	}
 	foreach my $close (@closes) {
-	    if (exists $statuses->{$close}) {
+	    if ($statuses and exists $statuses->{$close}) {
 		my $title = $statuses->{$close}->{subject};
 		my $pkg = $statuses->{$close}->{package};
 		$title =~ s/^($pkg|$PACKAGE): //;
-		push @closes_text, "$title (Closes: \#$close)\n";
+		push @closes_text, "Fix \"$title\" <explain what you changed and why> (Closes: \#$close)\n";
 	    }
 	    else { # not our package, or wnpp
 		my $bug = Devscripts::Debbugs::status(
 		    [bug => $close, indicatesource => 1] );
-		if (not defined $bug) {
+		if ($bug eq "") {
 		    warn "$progname warning: unknown bug \#$close does not belong to $PACKAGE,\n  disabling closing changelog entry\n";
 		    $warnings++;
 		    push @closes_text, "Closes?? \#$close: UNKNOWN BUG IN WRONG PACKAGE!!\n";
@@ -827,7 +864,7 @@ my $date_cmd = ($opt_tz ? "TZ=$opt_tz " : "") . "date -R";
 chomp(my $DATE=`$date_cmd`);
 
 if ($opt_news && !$opt_i && !$opt_a) {
-    if ($VERSION eq $changelog{'Version'}) {
+    if ($VERSION eq $changelog{'Version'} && !$opt_v && !$opt_l) {
 	$opt_a = 1;
     } else {
 	$opt_i = 1;
@@ -836,7 +873,7 @@ if ($opt_news && !$opt_i && !$opt_a) {
 
 # Are we going to have to figure things out for ourselves?
 if (! $opt_i && ! $opt_v && ! $opt_d && ! $opt_a && ! $opt_e && ! $opt_r &&
-    ! $opt_n && ! $opt_bn && ! $opt_qa && ! $opt_s && ! $opt_bpo &&
+    ! $opt_n && ! $opt_bn && ! $opt_R && ! $opt_qa && ! $opt_s && ! $opt_bpo &&
     ! $opt_l && ! $opt_create) {
     # Yes, we are
     if ($opt_release_heuristic eq 'log') {
@@ -913,7 +950,7 @@ my $line;
 my $optionsok=0;
 my $merge=0;
 
-if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_s || $opt_bpo || $opt_l || $opt_v || $opt_d ||
+if (($opt_i || $opt_n || $opt_bn || $opt_R || $opt_qa || $opt_s || $opt_bpo || $opt_l || $opt_v || $opt_d ||
     ($opt_news && $VERSION ne $changelog{'Version'})) && ! $opt_create) {
 
     $optionsok=1;
@@ -957,7 +994,7 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_s || $opt_bpo || $opt_l || $
 	    $opt_p=1;
 	}
 
-	if (system("dpkg --compare-versions $VERSION lt $NEW_VERSION" .
+	if (system("dpkg --compare-versions $VERSION le $NEW_VERSION" .
 		  " 2>/dev/null 1>&2")) {
 	    if ($opt_b or ($opt_allow_lower and $NEW_VERSION =~ /$opt_allow_lower/)) {
 		warn "$progname warning: new version ($NEW_VERSION) is less than\n" .
@@ -979,19 +1016,15 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_s || $opt_bpo || $opt_l || $
     # including epochs.
 
     if (! $NEW_VERSION) {
-	if ($VERSION =~ /(.*?)([a-yA-Y][a-zA-Z]*|\d+)$/i) {
+	if ($VERSION =~ /(.*?)([a-yA-Y][a-zA-Z]*|\d+)(~)?$/i) {
+	    my $extra=$3 || '';
+	    my $useextra = 0;
 	    my $end=$2;
 	    my $start=$1;
 	    # If it's not already an NMU make it so
 	    # otherwise we can be safe if we behave like dch -i
 
-	    if ($distributor eq 'Ubuntu' and $start !~ /(build|ubuntu|prop(osed)?)(\d+\.)*$/ and not $opt_U) {
-		if ($start =~ /~ppa([0-9]*)$/) {
-		    $end++;
-		} else {
-		    $end .= "ubuntu1";
-		}
-	    } elsif (($opt_n or $opt_s) and (
+	    if (($opt_n or $opt_s) and $distributor ne 'Ubuntu' and (
 		($VERSION eq $UVERSION and not $start =~ /\+nmu/)
 		or ($VERSION ne $UVERSION and not $start =~ /\.$/))) {
 
@@ -999,21 +1032,23 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_s || $opt_bpo || $opt_l || $
 		    # First NMU of a Debian native package
 		    $end .= "+nmu1";
 		} else {
-	    	    $end += 0.1;
+		    $end += 0.1;
 		}
-            } elsif ($opt_bn and not $start =~ /\+b/) {
-                $end .= "+b1";
+	    } elsif ($opt_bn and not $start =~ /\+b/) {
+		$end .= "+b1";
+	    } elsif ($opt_R and not $start =~ /build/ and not $start =~ /ubuntu/) {
+		$end .= "build1";
 	    } elsif ($opt_qa and $start =~/(.*?)-(\d+)\.$/) {
-                # Drop NMU revision when doing a QA upload
-                my $upstream_version = $1;
-                my $debian_revision = $2;
-                $debian_revision++;
-                $start = "$upstream_version-$debian_revision";
-                $end = "";
-	    } elsif ($opt_bpo and not $start =~ /~bpo\.$/) {
+		# Drop NMU revision when doing a QA upload
+		my $upstream_version = $1;
+		my $debian_revision = $2;
+		$debian_revision++;
+		$start = "$upstream_version-$debian_revision";
+		$end = "";
+	    } elsif ($opt_bpo and not $start =~ /~bpo[0-9]+\+$/) {
 		# If it's not already a backport make it so
 		# otherwise we can be safe if we behave like dch -i
-		$end .= "~bpo40+1";
+		$end .= "~bpo$latest_bpo_dist+1";
 	    } elsif ($opt_l and not $start =~ /\Q$opt_l\E/) {
 		# If it's not already a local package make it so
 		# otherwise we can be safe if we behave like dch -i
@@ -1021,10 +1056,49 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_s || $opt_bpo || $opt_l || $
 	    } elsif (!$opt_news) {
 		# Don't bump the version of a NEWS file in this case as we're
 		# using the version from the changelog
-		$end++;
+
+		if (($opt_i or $opt_s) and $distributor eq 'Ubuntu' and
+		    $start !~ /(build|ubuntu|prop(osed)?|~ppa)(\d+\.)*$/ and
+		    not $opt_U) {
+		    $end .= "ubuntu1";
+		} else {
+		    $end++;
+		}
+
+		# Attempt to set the distribution for a backport correctly
+		# based on the version of the previous backport
+		if ($opt_bpo) {
+		    my $previous_dist = $start;
+		    $previous_dist =~ s/^.*~bpo([0-9]+)\+$/$1/;
+		    if (defined $previous_dist and defined
+			$bpo_dists{$previous_dist}) {
+			$bpo_dist = $bpo_dists{$previous_dist} . '-backports';
+		    } else {
+			# Fallback to using the previous distribution
+			$bpo_dist = $changelog{'Distribution'};
+		    }
+		}
+
+		if(! ($opt_s or $opt_n or $distributor eq 'Ubuntu')) {
+		    if ($start =~/(.*?)-(\d+)\.$/) {
+			# Drop NMU revision
+			my $upstream_version = $1;
+			my $debian_revision = $2;
+			$debian_revision++;
+			$start = "$upstream_version-$debian_revision";
+			$end = "";
+		    }
+		}
+
+		if (! ($opt_qa or $opt_bpo or $opt_l)) {
+		    $useextra = 1;
+		}
 	    }
 	    $DISTRIBUTION = $changelog{'Distribution'} if $start =~ /prop(osed)?(\d+\.)*$/;
 	    $NEW_VERSION = "$start$end";
+	    if ($useextra) {
+		$NEW_VERSION .= $extra;
+	    }
 	    ($NEW_SVERSION=$NEW_VERSION) =~ s/^\d+://;
 	    ($NEW_UVERSION=$NEW_SVERSION) =~ s/-[^-]*$//;
 	} else {
@@ -1040,13 +1114,18 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_s || $opt_bpo || $opt_l || $
     }
 
     if ($opt_bpo) {
-	$bpo_dist='etch-backports';
+	$bpo_dist ||= $bpo_dists{$latest_bpo_dist} . '-backports';
     }
     my $distribution = $opt_D || $bpo_dist || (($opt_release_heuristic eq 'changelog') ? "UNRELEASED" : $DISTRIBUTION);
-    
-    my $urgency = $opt_u || 'low';
 
-    if (($opt_v or $opt_i or $opt_l) and $opt_release_heuristic eq "changelog" and
+    my $urgency = $opt_u;
+    if ($opt_news) {
+	$urgency ||= $CL_URGENCY;
+    }
+    $urgency ||= 'low';
+
+    if (($opt_v or $opt_i or $opt_l or $opt_d) and
+	$opt_release_heuristic eq "changelog" and
 	$changelog{'Distribution'} eq "UNRELEASED" and
 	$distribution eq "UNRELEASED") {
 
@@ -1054,15 +1133,15 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_s || $opt_bpo || $opt_l || $
     } else {
 	print O "$PACKAGE ($NEW_VERSION) $distribution; urgency=$urgency\n\n";
 	if ($opt_n && ! $opt_news) {
-            print O "  * Non-maintainer upload.\n";
+	    print O "  * Non-maintainer upload.\n";
 	    $line = 1;
 	} elsif ($opt_bn && ! $opt_news) {
-            my $arch = qx/dpkg-architecture -qDEB_BUILD_ARCH/; chomp ($arch);
-            print O "  * Binary-only non-maintainer upload for $arch; no source changes.\n";
-            $line = 1;
+	    my $arch = qx/dpkg-architecture -qDEB_BUILD_ARCH/; chomp ($arch);
+	    print O "  * Binary-only non-maintainer upload for $arch; no source changes.\n";
+	    $line = 1;
 	} elsif ($opt_qa && ! $opt_news) {
-            print O "  * QA upload.\n";
-            $line = 1;
+	    print O "  * QA upload.\n";
+	    $line = 1;
 	} elsif ($opt_s && ! $opt_news) {
 	    if ($distributor eq 'Ubuntu') {
 		print O "  * SECURITY UPDATE:\n";
@@ -1070,18 +1149,18 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_s || $opt_bpo || $opt_l || $
 	    } else {
 		print O "  * Non-maintainer upload by the Security Team.\n";
 	    }
-            $line = 1;
+	    $line = 1;
 	} elsif ($opt_bpo && ! $opt_news) {
-            print O "  * Rebuild for etch-backports.\n";
-            $line = 1;
+	    print O "  * Rebuild for $bpo_dist.\n";
+	    $line = 1;
 	}
 	if (@closes_text or $TEXT) {
-            foreach (@closes_text) { format_line($_, 1); }
-            if (length $TEXT) { format_line($TEXT, 1); }
+	    foreach (@closes_text) { format_line($_, 1); }
+	    if (length $TEXT) { format_line($TEXT, 1); }
 	} elsif ($opt_news) {
-            print O "  \n";
+	    print O "  \n";
 	} else {
-            print O "  * \n";
+	    print O "  * \n";
 	}
 	$line += 3;
 	print O "\n -- $MAINTAINER <$EMAIL>  $DATE\n\n";
@@ -1094,7 +1173,7 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_s || $opt_bpo || $opt_l || $
 if (($opt_r || $opt_a || $merge) && ! $opt_create) {
     # This means we just have to generate a new * entry in changelog
     # and if a multi-developer changelog is detected, add developer names.
-    
+
     $NEW_VERSION=$VERSION unless $NEW_VERSION;
     $NEW_SVERSION=$SVERSION unless $NEW_SVERSION;
     $NEW_UVERSION=$UVERSION unless $NEW_UVERSION;
@@ -1103,7 +1182,7 @@ if (($opt_r || $opt_a || $merge) && ! $opt_create) {
     # last entry, and determine whether there are existing
     # multi-developer changes by the current maintainer.
     $line=-1;
-    my ($lastmaint, $nextmaint, $maintline, $count, $lastheader, $lastdist);
+    my ($lastmaint, $nextmaint, $maintline, $count, $lastheader, $lastdist, $dist_indicator);
     my $savedline = $line;;
     while (<S>) {
 	$line++;
@@ -1128,7 +1207,11 @@ if (($opt_r || $opt_a || $merge) && ! $opt_create) {
 		$line = $savedline;
 		last;
 	    }
-	}	
+	}
+	elsif (/  \* Upload to ([^ ]+).*$/) {
+	    ($dist_indicator = $1) =~ s/[!:.,;]$//;
+	    chomp $dist_indicator;
+	}
 	elsif (/^ --\s+([^<]+)\s+/) {
 	    $lastmaint=$1;
 	    # Remember where we are so we can skip back afterwards
@@ -1173,13 +1256,22 @@ if (($opt_r || $opt_a || $merge) && ! $opt_create) {
     if ($CHANGES =~ m/^\w[-+0-9a-z.]* \([^\(\) \t]+\)((?:\s+[-+0-9a-z.]+)+)\;\s+urgency=(\w+)/i) {
 	my $distribution = $1;
 	my $urgency = $2;
+	if ($opt_news) {
+	    $urgency = $CL_URGENCY;
+	}
 	$distribution =~ s/^\s+//;
 	if ($opt_r) {
 	    # Change the distribution from UNRELEASED for release
-	    if ($distributor eq 'Ubuntu') {
-		$distribution = $opt_D || "intrepid";
-	    } elsif ($distribution eq "UNRELEASED") {
-		$distribution = $opt_D || $lastdist || "unstable";
+	    if ($distribution eq "UNRELEASED") {
+		if ($dist_indicator and not $opt_D) {
+		    $distribution = $dist_indicator;
+		} else {
+		    if ($distributor eq 'Ubuntu') {
+			$distribution = $opt_D || "lucid";
+		    } else {
+			$distribution = $opt_D || $lastdist || "unstable";
+		    }
+		}
 	    } elsif ($opt_D) {
 		warn "$progname warning: ignoring distribution passed to --release as changelog has already been released\n";
 	    }
@@ -1210,7 +1302,7 @@ if (($opt_r || $opt_a || $merge) && ! $opt_create) {
     };
 
     if (! $opt_r) {
-    	# Add a multi-maintainer header...
+	# Add a multi-maintainer header...
 	if ($multimaint) {
 	    # ...unless there already is one for this maintainer.
 	    if (!defined $maintline) {
@@ -1283,7 +1375,11 @@ elsif ($opt_create) {
 	push @closes_text, "Initial release. (Closes: \#XXXXXX)\n";
     }
 
-    my $urgency = $opt_u || 'low';
+    my $urgency = $opt_u;
+    if ($opt_news) {
+	$urgency ||= $CL_URGENCY;
+    }
+    $urgency ||= 'low';
     print O "$PACKAGE ($VERSION) $DISTRIBUTION; urgency=$urgency\n\n";
 
     if (@closes_text or $TEXT) {
@@ -1335,7 +1431,9 @@ if ((!$TEXT and !$EMPTY_TEXT and ! ($opt_create and $opt_empty)) or @closes_text
 	my $newmtime = (stat("$changelog_path.dch"))[9];
 	defined $newmtime or fatal
 	    "Error getting modification time of temporary $changelog_path: $!";
-	if ($mtime == $newmtime && ! $opt_r && ! $opt_create) {
+	if ($mtime == $newmtime && ! $opt_create &&
+	    (!$opt_r || ($opt_r && $opt_force_save_on_release))) {
+
 	    warn "$progname: $changelog_path unmodified; exiting.\n";
 	    exit 0;
 	}
@@ -1377,6 +1475,12 @@ if ((basename(cwd()) =~ m%^\Q$PACKAGE\E-\Q$UVERSION\E$%) &&
 	} else {
 	    warn "$progname warning: Couldn't rename directory: $!";
 	}
+	# And check whether a new orig tarball exists
+	my @origs = glob("../$PACKAGE\_$new_uversion.*");
+	my $num_origs = grep { /^..\/\Q$PACKAGE\E_\Q$new_uversion\E\.orig\.tar\.(gz|bz2|lzma)$/ } @origs;
+	if ($num_origs == 0) {
+	    warn "$progname warning: no orig tarball found for the new version.\n";
+	}
     }
 }
 
@@ -1404,10 +1508,13 @@ sub format_line {
     my $newentry=shift;
 
     # Work around the fact that write() with formats
-    # seems to assume that character == byte
+    # seems to assume that characters are single-byte
     # See http://rt.perl.org/rt3/Public/Bug/Display.html?id=33832
-    # and Debian bug #473769
-    my $count = () = $CHGLINE =~ /[^\x00-\x7F]/mg;
+    # and Debian bugs #473769 and #541484
+    # This relies on $CHGLINE being a sequence of unicode characters.  We can
+    # compare how many unicode characters we have to how many bytes we have
+    # when encoding to utf8 and therefore how many spaces we need to pad.
+    my $count = length(encode_utf8($CHGLINE)) - length($CHGLINE);
     $CHGLINE .= " " x $count;
 
     print O "\n" if $opt_news && ! ($newentry || $linecount);
