@@ -17,14 +17,16 @@
 use 5.006_000;
 use strict;
 use Cwd;
-use Dpkg::ErrorHandling qw / error internerr syserr subprocerr /;
-use Dpkg::Gettext;
+use Dpkg::IPC;
 use File::Basename;
 use File::Copy qw/ cp move /;
 use File::Path qw/ rmtree /;
-use File::Temp qw/ tempdir /;
+use File::Temp qw/ tempdir tempfile /;
+use lib '/usr/share/devscripts';
+use Devscripts::Versort;
 
 # Predeclare functions
+sub wdiff_control_files($$$$$);
 sub process_debc($$);
 sub process_debI($);
 sub mktmpdirs();
@@ -33,195 +35,7 @@ sub fatal(@);
 my $progname = basename($0);
 my $modified_conf_msg;
 my $exit_status = 0;
-
-# _sanity_check_opts(), _fork_and_exec(), and _wait_child() were pulled from
-# Lucid's Dpkg::IPC since a security update required spawn() to be available.
-sub _sanity_check_opts {
-    my (%opts) = @_;
-
-    internerr("exec parameter is mandatory in fork_and_exec()")
-	unless $opts{"exec"};
-
-    my $to = my $error_to = my $from = 0;
-    foreach (qw(file handle string pipe)) {
-	$to++ if $opts{"to_$_"};
-	$error_to++ if $opts{"error_to_$_"};
-	$from++ if $opts{"from_$_"};
-    }
-    internerr("not more than one of to_* parameters is allowed")
-	if $to > 1;
-    internerr("not more than one of error_to_* parameters is allowed")
-	if $error_to > 1;
-    internerr("not more than one of from_* parameters is allowed")
-	if $from > 1;
-
-    foreach (qw(to_string error_to_string from_string
-                to_pipe error_to_pipe from_pipe))
-    {
-	if (exists $opts{$_} and
-	    (!ref($opts{$_}) or ref($opts{$_}) ne 'SCALAR')) {
-	    internerr("parameter $_ must be a scalar reference");
-	}
-    }
-
-    if (exists $opts{"timeout"} and defined($opts{"timeout"}) and
-        $opts{"timeout"} !~ /^\d+$/) {
-	internerr("parameter timeout must be an integer");
-    }
-
-    if (exists $opts{"env"} and ref($opts{"env"}) ne 'HASH') {
-	internerr("parameter env must be a hash reference");
-    }
-
-    if (exists $opts{"delete_env"} and ref($opts{"delete_env"}) ne 'ARRAY') {
-	internerr("parameter delete_env must be an array reference");
-    }
-
-    return %opts;
-}
-
-sub _fork_and_exec {
-    my (%opts) = _sanity_check_opts(@_);
-    $opts{"close_in_child"} ||= [];
-    my @prog;
-    if (ref($opts{"exec"}) =~ /ARRAY/) {
-	push @prog, @{$opts{"exec"}};
-    } elsif (not ref($opts{"exec"})) {
-	push @prog, $opts{"exec"};
-    } else {
-	internerr("invalid exec parameter in fork_and_exec()");
-    }
-    my ($from_string_pipe, $to_string_pipe, $error_to_string_pipe);
-    if ($opts{"to_string"}) {
-	$opts{"to_pipe"} = \$to_string_pipe;
-	$opts{"wait_child"} = 1;
-    }
-    if ($opts{"error_to_string"}) {
-	$opts{"error_to_pipe"} = \$error_to_string_pipe;
-	$opts{"wait_child"} = 1;
-    }
-    if ($opts{"from_string"}) {
-	$opts{"from_pipe"} = \$from_string_pipe;
-    }
-    # Create pipes if needed
-    my ($input_pipe, $output_pipe, $error_pipe);
-    if ($opts{"from_pipe"}) {
-	pipe($opts{"from_handle"}, $input_pipe) ||
-		syserr(_g("pipe for %s"), "@prog");
-	${$opts{"from_pipe"}} = $input_pipe;
-	push @{$opts{"close_in_child"}}, $input_pipe;
-    }
-    if ($opts{"to_pipe"}) {
-	pipe($output_pipe, $opts{"to_handle"}) ||
-		syserr(_g("pipe for %s"), "@prog");
-	${$opts{"to_pipe"}} = $output_pipe;
-	push @{$opts{"close_in_child"}}, $output_pipe;
-    }
-    if ($opts{"error_to_pipe"}) {
-	pipe($error_pipe, $opts{"error_to_handle"}) ||
-		syserr(_g("pipe for %s"), "@prog");
-	${$opts{"error_to_pipe"}} = $error_pipe;
-	push @{$opts{"close_in_child"}}, $error_pipe;
-    }
-    # Fork and exec
-    my $pid = fork();
-    syserr(_g("fork for %s"), "@prog") unless defined $pid;
-    if (not $pid) {
-	# Define environment variables
-	if ($opts{"env"}) {
-	    foreach (keys %{$opts{"env"}}) {
-		$ENV{$_} = $opts{"env"}{$_};
-	    }
-	}
-	if ($opts{"delete_env"}) {
-	    delete $ENV{$_} foreach (@{$opts{"delete_env"}});
-	}
-	# Change the current directory
-	if ($opts{"chdir"}) {
-	    chdir($opts{"chdir"}) || syserr(_g("chdir to %s"), $opts{"chdir"});
-	}
-	# Redirect STDIN if needed
-	if ($opts{"from_file"}) {
-	    open(STDIN, "<", $opts{"from_file"}) ||
-		syserr(_g("cannot open %s"), $opts{"from_file"});
-	} elsif ($opts{"from_handle"}) {
-	    open(STDIN, "<&", $opts{"from_handle"}) || syserr(_g("reopen stdin"));
-	    close($opts{"from_handle"}); # has been duped, can be closed
-	}
-	# Redirect STDOUT if needed
-	if ($opts{"to_file"}) {
-	    open(STDOUT, ">", $opts{"to_file"}) ||
-		syserr(_g("cannot write %s"), $opts{"to_file"});
-	} elsif ($opts{"to_handle"}) {
-	    open(STDOUT, ">&", $opts{"to_handle"}) || syserr(_g("reopen stdout"));
-	    close($opts{"to_handle"}); # has been duped, can be closed
-	}
-	# Redirect STDERR if needed
-	if ($opts{"error_to_file"}) {
-	    open(STDERR, ">", $opts{"error_to_file"}) ||
-		syserr(_g("cannot write %s"), $opts{"error_to_file"});
-	} elsif ($opts{"error_to_handle"}) {
-	    open(STDERR, ">&", $opts{"error_to_handle"}) || syserr(_g("reopen stdout"));
-	    close($opts{"error_to_handle"}); # has been duped, can be closed
-	}
-	# Close some inherited filehandles
-	close($_) foreach (@{$opts{"close_in_child"}});
-	# Execute the program
-	exec({ $prog[0] } @prog) or syserr(_g("exec %s"), "@prog");
-    }
-    # Close handle that we can't use any more
-    close($opts{"from_handle"}) if exists $opts{"from_handle"};
-    close($opts{"to_handle"}) if exists $opts{"to_handle"};
-    close($opts{"error_to_handle"}) if exists $opts{"error_to_handle"};
-
-    if ($opts{"from_string"}) {
-	print $from_string_pipe ${$opts{"from_string"}};
-	close($from_string_pipe);
-    }
-    if ($opts{"to_string"}) {
-	local $/ = undef;
-	${$opts{"to_string"}} = readline($to_string_pipe);
-    }
-    if ($opts{"error_to_string"}) {
-	local $/ = undef;
-	${$opts{"error_to_string"}} = readline($error_to_string_pipe);
-    }
-    if ($opts{"wait_child"}) {
-	my $cmdline = "@prog";
-	if ($opts{"env"}) {
-	    foreach (keys %{$opts{"env"}}) {
-		$cmdline = "$_=\"" . $opts{"env"}{$_} . "\" $cmdline";
-	    }
-	}
-	_wait_child($pid, nocheck => $opts{"nocheck"},
-                   timeout => $opts{"timeout"}, cmdline => $cmdline);
-	return 1;
-    }
-
-    return $pid;
-}
-
-sub _wait_child {
-    my ($pid, %opts) = @_;
-    $opts{"cmdline"} ||= _g("child process");
-    internerr("no PID set, cannot wait end of process") unless $pid;
-    eval {
-        local $SIG{ALRM} = sub { die "alarm\n" };
-        alarm($opts{"timeout"}) if defined($opts{"timeout"});
-        $pid == waitpid($pid, 0) or syserr(_g("wait for %s"), $opts{"cmdline"});
-        alarm(0) if defined($opts{"timeout"});
-    };
-    if ($@) {
-        die $@ unless $@ eq "alarm\n";
-        error(ngettext("%s didn't complete in %d second",
-                       "%s didn't complete in %d seconds",
-                       $opts{"timeout"}),
-              $opts{"cmdline"}, $opts{"timeout"});
-    }
-    unless ($opts{"nocheck"}) {
-	subprocerr($opts{"cmdline"}) if $?;
-    }
-}
+my $dummyname = "---DUMMY---";
 
 sub usage {
     print <<"EOF";
@@ -252,6 +66,10 @@ Valid options are:
                             ALL to compare all control files present
    --wp, --wl, --wt       Pass the option -p, -l, -t respectively to wdiff
                             (only one should be used)
+   --wdiff-source-control When processing source packages, compare control
+                            files as with --control for binary packages
+   --no-wdiff-source-control
+                          Do not do so (default)
    --show-moved           Indicate also all files which have moved
                             between packages
    --noshow-moved         Do not also indicate all files which have moved
@@ -262,6 +80,14 @@ Valid options are:
    --quiet, -q            Be quiet if no differences were found
    --exclude PATTERN      Exclude files that match PATTERN
    --ignore-space, -w     Ignore whitespace in diffs
+   --diffstat             Include the result of diffstat before the diff
+   --no-diffstat          Do not do so (default)
+   --auto-ver-sort        When comparing source packages, ensure the
+                          comparison is performed in version order
+   --no-auto-ver-sort     Do not do so (default)
+   --unpack-tarballs      Unpack tarballs found in the top level source
+                          directory (default)
+   --no-unpack-tarballs   Do not do so
 
 Default settings modified by devscripts configuration files:
 $modified_conf_msg
@@ -286,6 +112,10 @@ my $controlfiles = 'control';
 my $show_moved = 0;
 my $wdiff_opt = '';
 my @diff_opts = ();
+my $show_diffstat = 0;
+my $wdiff_source_control = 0;
+my $auto_ver_sort = 0;
+my $unpack_tarballs = 1;
 
 my $quiet = 0;
 
@@ -303,6 +133,10 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 		       'DEBDIFF_CONTROLFILES' => 'control',
 		       'DEBDIFF_SHOW_MOVED' => 'no',
 		       'DEBDIFF_WDIFF_OPT' => '',
+		       'DEBDIFF_SHOW_DIFFSTAT' => 'no',
+		       'DEBDIFF_WDIFF_SOURCE_CONTROL' => 'no',
+		       'DEBDIFF_AUTO_VER_SORT' => 'no',
+		       'DEBDIFF_UNPACK_TARBALLS' => 'yes',
 		       );
     my %config_default = %config_vars;
 
@@ -325,6 +159,14 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 	or $config_vars{'DEBDIFF_CONTROL'}='yes';
     $config_vars{'DEBDIFF_SHOW_MOVED'} =~ /^(yes|no)$/
 	or $config_vars{'DEBDIFF_SHOW_MOVED'}='no';
+    $config_vars{'DEBDIFF_SHOW_DIFFSTAT'} =~ /^(yes|no)$/
+	or $config_vars{'DEBDIFF_SHOW_DIFFSTAT'}='no';
+    $config_vars{'DEBDIFF_WDIFF_SOURCE_CONTROL'} =~ /^(yes|no)$/
+	or $config_vars{'DEBDIFF_WDIFF_SOURCE_CONTROL'}='no';
+    $config_vars{'DEBDIFF_AUTO_VER_SORT'} =~ /^(yes|no)$/
+	or $config_vars{'DEBDIFF_AUTO_VER_SORT'}='no';
+    $config_vars{'DEBDIFF_UNPACK_TARBALLS'} =~ /^(yes|no)$/
+	or $config_vars{'DEBDIFF_UNPACK_TARBALLS'}='yes';
 
     foreach my $var (sort keys %config_vars) {
 	if ($config_vars{$var} ne $config_default{$var}) {
@@ -339,6 +181,12 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     $controlfiles = $config_vars{'DEBDIFF_CONTROLFILES'};
     $show_moved = $config_vars{'DEBDIFF_SHOW_MOVED'} eq 'yes' ? 1 : 0;
     $wdiff_opt = $config_vars{'DEBDIFF_WDIFF_OPT'} =~ /^-([plt])$/ ? $1 : '';
+    $show_diffstat = $config_vars{'DEBDIFF_SHOW_DIFFSTAT'} eq 'yes' ? 1 : 0;
+    $wdiff_source_control = $config_vars{'DEBDIFF_WDIFF_SOURCE_CONTROL'}
+	eq 'yes' ? 1 : 0;
+    $auto_ver_sort = $config_vars{'DEBDIFF_AUTO_VER_SORT'} eq 'yes' ? 1 : 0;
+    $unpack_tarballs = $config_vars{'DEBDIFF_UNPACK_TARBALLS'} eq 'yes' ? 1 : 0;
+
 }
 
 # Are they a pair of debs, changes or dsc files, or a list of debs?
@@ -419,6 +267,14 @@ while (@ARGV) {
 	push @diff_opts, "-w"; 
 	shift;
     }
+    elsif ($ARGV[0] eq '--diffstat') { $show_diffstat = 1; shift; }
+    elsif ($ARGV[0] =~ /^--no-?diffstat$/) { $show_diffstat = 0; shift; }
+    elsif ($ARGV[0] eq '--wdiff-source-control') { $wdiff_source_control = 1; shift; }
+    elsif ($ARGV[0] =~ /^--no-?wdiff-source-control$/) { $wdiff_source_control = 0; shift; }
+    elsif ($ARGV[0] eq '--auto-ver-sort') { $auto_ver_sort = 1; shift; }
+    elsif ($ARGV[0] =~ /^--no-?auto-ver-sort$/) { $auto_ver_sort = 0; shift; }
+    elsif ($ARGV[0] eq '--unpack-tarballs') { $unpack_tarballs = 1; shift; }
+    elsif ($ARGV[0] =~ /^--no-?unpack-tarballs$/) { $unpack_tarballs = 0; shift; }
     elsif ($ARGV[0] =~ /^--no-?conf$/) {
 	fatal "--no-conf is only acceptable as the first command-line option!";
     }
@@ -438,10 +294,12 @@ my $guessed_version = 0;
 # If no file is given, assume that we are in a source directory
 # and try to create a diff with the previous version
 if(@ARGV == 0) {
+    my $namepat = qr/[-+0-9a-z.]/i;
+
     fatal "Can't read file: debian/changelog" unless -r "debian/changelog";
     open CHL, "debian/changelog";
     while(<CHL>) {
-	if(/^(.+)\s\((\d+:)?(.+)\)\s(\w+)\;\surgency=.+$/) {
+	if(/^(\w$namepat*)\s\((\d+:)?(.+)\)((\s+$namepat+)+)\;\surgency=.+$/) {
 	    unshift @ARGV, "../".$1."_".$3.".dsc";
 	    $guessed_version++;
 	}
@@ -486,7 +344,7 @@ if ($type eq 'deb') {
 	my ($debc, $debI) = ('', '');
 	my %dpkg_env = ( LC_ALL => 'C' );
 	eval {
-	    _fork_and_exec(exec => ['dpkg-deb', '-c', $deb],
+	    fork_and_exec(exec => ['dpkg-deb', '-c', $deb],
 		env => \%dpkg_env,
 		to_string => \$debc,
 		wait_child => 1);
@@ -496,7 +354,7 @@ if ($type eq 'deb') {
 	}
 
 	eval {
-	    _fork_and_exec(exec => ['dpkg-deb', '-I', $deb],
+	    fork_and_exec(exec => ['dpkg-deb', '-I', $deb],
 		env => \%dpkg_env,
 		to_string => \$debI,
 		wait_child => 1);
@@ -562,7 +420,7 @@ elsif ($type eq 'changes' or $type eq 'debs') {
 	    my ($debc, $debI) = ('', '');
 	    my %dpkg_env = ( LC_ALL => 'C' );
 	    eval {
-		_fork_and_exec(exec => ['dpkg-deb', '-c', $deb],
+		fork_and_exec(exec => ['dpkg-deb', '-c', $deb],
 		    to_string => \$debc,
 		    env => \%dpkg_env,
 		    wait_child => 1);
@@ -571,7 +429,7 @@ elsif ($type eq 'changes' or $type eq 'debs') {
 		fatal "dpkg-deb -c $deb failed!";
 	    }
 	    eval {
-		_fork_and_exec(exec => ['dpkg-deb', '-I', $deb],
+		fork_and_exec(exec => ['dpkg-deb', '-I', $deb],
 		    to_string => \$debI,
 		    env => \%dpkg_env,
 		    wait_child => 1);
@@ -608,7 +466,7 @@ elsif ($type eq 'dsc') {
     # Compare source packages
     my $pwd = cwd;
 
-    my (@origs, @diffs, @dscs);
+    my (@origs, @diffs, @dscs, @dscformats, @versions);
     foreach my $i (1,2) {
 	my $dsc = shift;
 	chdir dirname($dsc)
@@ -623,6 +481,10 @@ elsif ($type eq 'dsc') {
 	    if (/^Files:/) {
 		$infiles=1;
 		next;
+	    } elsif (/^Format: (.*)$/) {
+		$dscformats[$i] = $1;
+	    } elsif (/^Version: (.*)$/) {
+		$versions[$i - 1] = [ $1, $i ];
 	    }
 	    next unless $infiles;
 	    last if /^\s*$/;
@@ -637,7 +499,7 @@ elsif ($type eq 'dsc') {
 		if ($file =~ /\.diff\.gz$/) {
 		    $diffs[$i] = cwd() . '/' . $file;
 		}
-		elsif ($file =~ /(\.orig)?\.tar\.gz$/) {
+		elsif ($file =~ /(\.orig)?\.tar\.(gz|bz2|lzma)$/) {
 		    $origs[$i] = $file;
 		}
 	    } else {
@@ -650,35 +512,77 @@ elsif ($type eq 'dsc') {
 	chdir $pwd or fatal "Couldn't chdir $pwd: $!";
     }
 
+    @versions = Devscripts::Versort::versort(@versions);
+    # If the versions are currently out of order, should we swap them?
+    if ($auto_ver_sort and !$guessed_version and $versions[0][1] == 1) {
+	foreach my $var ((\@origs, \@diffs, \@dscs, \@dscformats)) {
+	    my $temp = @{$var}[1];
+	    @{$var}[1] = @{$var}[2];
+	    @{$var}[2] = $temp;
+	}
+    }
+
     # Do we have interdiff?
     system("command -v interdiff >/dev/null 2>&1");
     my $use_interdiff = ($?==0) ? 1 : 0;
+    system("command -v diffstat >/dev/null 2>&1");
+    my $have_diffstat = ($?==0) ? 1 : 0;
+    system("command -v wdiff >/dev/null 2>&1");
+    my $have_wdiff = ($?==0) ? 1 : 0;
 
+    my ($fh, $filename) = tempfile("debdiffXXXXXX",
+				SUFFIX => ".diff",
+				DIR => File::Spec->tmpdir,
+				UNLINK => 1);
+
+    # When wdiffing source control files we always fully extract both source
+    # packages as it's the easiest way of getting the debian/control file,
+    # particularly if the orig tar ball contains one which is patched in the
+    # diffs
     if ($origs[1] eq $origs[2] and defined $diffs[1] and defined $diffs[2]
-	and scalar(@excludes) == 0 and $use_interdiff) {
-	# same orig tar ball and interdiff exists
-	eval {
-	    _fork_and_exec(exec => ['interdiff', '-z', @diff_opts, $diffs[1], $diffs[2]],
-		  wait_child => 1);
-	};
-	if ($@) {
-	    fatal "interdiff -z $diffs[1] $diffs[2] failed!";
+	and scalar(@excludes) == 0 and $use_interdiff and !$wdiff_source_control) {
+	# same orig tar ball, interdiff exists and not wdiffing
+
+	fork_and_exec(exec => ['interdiff', '-z', @diff_opts, $diffs[1], $diffs[2]],
+	      to_file => $filename,
+	      wait_child => 1);
+	if ($have_diffstat and $show_diffstat) {
+	    my $header = "diffstat for " . basename($diffs[1])
+			    . " " . basename($diffs[2]) . "\n\n";
+	    $header =~ s/\.diff\.gz//g;
+	    print $header;
+	    fork_and_exec(exec => ['diffstat', $filename],
+		wait_child => 1);
+	    print "\n";
+	}
+
+	if (-s $filename) {
+	    open( INTERDIFF, '<', $filename );
+	    while( <INTERDIFF> ) {
+		print $_;
+	    }
+	    close INTERDIFF;
+
+	    $exit_status = 1;
 	}
     } else {
 	# Any other situation
 	if ($origs[1] eq $origs[2] and
 	    defined $diffs[1] and defined $diffs[2] and
-	    scalar(@excludes) == 0) {
+	    scalar(@excludes) == 0 and !$wdiff_source_control) {
 	    warn "Warning: You do not seem to have interdiff (in the patchutils package)\ninstalled; this program would use it if it were available.\n";
 	}
-	# possibly different orig tarballs, or no interdiff installed
+	# possibly different orig tarballs, or no interdiff installed,
+	# or wdiffing debian/control
 	our ($sdir1, $sdir2);
 	mktmpdirs();
 	for my $i (1,2) {
 	    no strict 'refs';
+	    my @opts = ('-x');
+	    push (@opts, '--skip-patches') if $dscformats[$i] eq '3.0 (quilt)';
 	    my $diri = ${"dir$i"};
 	    eval {
-		_fork_and_exec(exec => ['dpkg-source', '-x', $dscs[$i]],
+		fork_and_exec(exec => ['dpkg-source', @opts, $dscs[$i]],
 		      to_file => '/dev/null',
 		      chdir => $diri,
 		      wait_child => 1);
@@ -690,7 +594,7 @@ elsif ($type eq 'dsc') {
 		my $dscx = basename $dscs[$i];
 		cp $diffs[$i], $diri || fatal "copy $diffs[$i] $diri: $!";
 		cp $dscs[$i], $diri || fatal "copy $dscs[$i] $diri: $!";
-		_fork_and_exec(exec => ['dpkg-source', '-x', $dscx],
+		fork_and_exec(exec => ['dpkg-source', @opts, $dscx],
 		      to_file => '/dev/null',
 		      chdir => $diri,
 		      wait_child => 1);
@@ -703,30 +607,45 @@ elsif ($type eq 'dsc') {
 	    }
 	    closedir(DIR);
 	    my $sdiri = ${"sdir$i"};
+
+	    # also unpack tarballs found in the top level source directory so we can compare their contents too
+	    next unless $unpack_tarballs;
 	    opendir DIR,$diri.'/'.$sdiri;
+
+	    my $tarballs = 1;
 	    while ($_ = readdir(DIR)) {
+		    my $unpacked = "=unpacked-tar" . $tarballs . "=";
+		    my $filename = $_;
 		    my $found = 0;
 		    my $comp = "";
 
 		    if ($_ =~ /tar.gz$/) {
+			$filename =~ s%(.*)\.tar\.gz$%$1%;
+			$tarballs++;
 			$found = 1;
 			$comp = "gzip";
 		    }
-		    if ($_ =~ /tar.bz$/) {
+		    if ($_ =~ /tar.bz$/ || $_ =~ /tar.bz2$/) {
+			$filename =~ s%(.*)\.tar\.bz2?$%$1%;
+			$tarballs++;
 			$found = 1;
 			$comp = "bzip2";
 		    }
 
 		    if ($found) {
-			_fork_and_exec(exec => ['tar', "--$comp", '-xf', $_],
+			fork_and_exec(exec => ['tar', "--$comp", '-xf', $_],
 			    to_file => '/dev/null',
 			    wait_child => 1,
 			    chdir => "$diri/$sdiri",
 			    nocheck => 1);
+			if (-d "$diri/$sdiri/$filename") {
+			    move "$diri/$sdiri/$filename", "$diri/$sdiri/$unpacked";
+			}
 		    }
 	    }
 	    closedir(DIR);
 	}
+
 	my @command = ("diff", "-Nru", @diff_opts);
 	for my $exclude (@excludes) {
 	    push @command, ("--exclude", $exclude);
@@ -735,23 +654,69 @@ elsif ($type eq 'dsc') {
 
 	# Execute diff and remove the common prefixes $dir1/$dir2, so the patch can be used with -p1,
 	# as if when interdiff would have been used:
-	open( DIFF, '-|', @command ) || fatal "Failed to execute @command!";
+	fork_and_exec(exec => \@command, to_file => $filename, wait_child => 1, nocheck => 1);
 
-	# replace in first line:
-	my $first = <DIFF>;
-	$first =~ s/ $dir1\/$sdir1/ $sdir1/;
-	$first =~ s/ $dir2\/$sdir2/ $sdir2/;
-	print $first;
+	if ($have_diffstat and $show_diffstat) {
+	    print "diffstat for $sdir1 $sdir2\n\n";
+	    fork_and_exec(exec => ['diffstat', $filename],
+		wait_child => 1);
+	    print "\n";
+	}
 
-	while(<DIFF>) {
+	if ($have_wdiff and $wdiff_source_control) {
+	    # Abuse global variables slightly to create some temporary directories
+	    my $tempdir1 = $dir1;
+	    my $tempdir2 = $dir2;
+	    mktmpdirs();
+	    our $wdiffdir1 = $dir1;
+	    our $wdiffdir2 = $dir2;
+	    $dir1 = $tempdir1;
+	    $dir2 = $tempdir2;
+	    our @cf;
+	    if ($controlfiles eq 'ALL') {
+		@cf = ('control');
+	    } else {
+		@cf = split /,/, $controlfiles;
+	    }
+
+	    no strict 'refs';
+	    for my $i (1,2) {
+		foreach my $file (@cf) {
+		    cp ${"dir$i"}.'/'.${"sdir$i"}."/debian/$file", ${"wdiffdir$i"};
+		}
+	    }
+	    use strict 'refs';
+
+	    # We don't support "ALL" for source packages as that would
+	    # wdiff debian/*
+	    $exit_status = wdiff_control_files($wdiffdir1, $wdiffdir2, $dummyname,
+		$controlfiles eq 'ALL' ? 'control' : $controlfiles,
+		$exit_status);
+	    print "\n";
+
+	    # Clean up
+	    rmtree([$wdiffdir1, $wdiffdir2]);
+	}
+
+	if (! -f $filename) {
+	    fatal "Creation of diff file $filename failed!";
+	} elsif (-s $filename) {
+	    open( DIFF, '<', $filename ) or fatal "Opening diff file $filename failed!";
+
+	    while(<DIFF>) {
 		s/^--- $dir1\//--- /;
 		s/^\+\+\+ $dir2\//+++ /;
+		s/^(diff .*) $dir1\/\Q$sdir1\E/$1 $sdir1/;
+		s/^(diff .*) $dir2\/\Q$sdir2\E/$1 $sdir2/;
 		print;
- 	}
-	close DIFF;
+	    }
+	    close DIFF;
+
+	    $exit_status = 1;
+	}
     }
 
-    exit 0;
+    exit $exit_status;
 }
 else {
     fatal "Internal error: \$type = $type unrecognised";
@@ -768,8 +733,8 @@ my (@deblosses, @debgains);
     grep $debs{$_}--, keys %debs1;
     grep $debs{$_}++, keys %debs2;
 
-    my @deblosses = sort grep $debs{$_} < 0, keys %debs;
-    my @debgains  = sort grep $debs{$_} > 0, keys %debs;
+    @deblosses = sort grep $debs{$_} < 0, keys %debs;
+    @debgains  = sort grep $debs{$_} > 0, keys %debs;
     @CommonDebs= sort grep $debs{$_} == 0, keys %debs;
 }
 
@@ -925,7 +890,6 @@ if ($show_moved and $type ne 'deb') {
 }
 
 # We compare the control files (at least the dependency fields)
-my $dummyname = "---DUMMY---";
 if (defined $singledeb[1] and defined $singledeb[2]) {
 	@CommonDebs = ( $dummyname );
 	$DebPaths1{$dummyname} = $singledeb[1];
@@ -947,7 +911,7 @@ for my $debname (@CommonDebs) {
 	my $debpath = "${\"DebPaths$i\"}{$debname}";
 	my $diri = ${"dir$i"};
 	eval {
-	    _fork_and_exec(exec => ['dpkg-deb', '-e', $debpath, $diri],
+	    fork_and_exec(exec => ['dpkg-deb', '-e', $debpath, $diri],
 		wait_child => 1);
 	};
 	if ($@) {
@@ -958,68 +922,9 @@ for my $debname (@CommonDebs) {
     }
 
     use strict 'refs';
+    $exit_status = wdiff_control_files($dir1, $dir2, $debname, $controlfiles,
+	$exit_status);
 
-    my @cf;
-    if ($controlfiles eq 'ALL') {
-	# only need to list one directory as we are only comparing control
-	# files in both packages
-	@cf = grep { ! /md5sums/ } map { basename($_); } glob("$dir1/*");
-    } else {
-	@cf = split /,/, $controlfiles;
-    }
-
-    foreach my $cf (@cf) {
-	next unless -f "$dir1/$cf" and -f "$dir2/$cf";
-	if ($cf eq 'control' or $cf eq 'conffiles') {
-	    for my $file ("$dir1/$cf", "$dir2/$cf") {
-		my ($fd, @hdrs);
-		open $fd, '<', $file or fatal "Cannot read $file: $!";
-		while (<$fd>) {
-		    if (/^\s/ and @hdrs > 0) {
-			$hdrs[$#hdrs] .= $_;
-		    } else {
-			push @hdrs, $_;
-		    }
-		}
-		close $fd;
-		open $fd, '>', $file or fatal "Cannot write $file: $!";
-		print $fd sort @hdrs;
-		close $fd;
-	    }
-	}
-	my $usepkgname = $debname eq $dummyname ? "" : " of package $debname";
-	my @opts = ('-n');
-	push @opts, $wdiff_opt if $wdiff_opt;
-	my $wdiff = '';
-	eval {
-	    _fork_and_exec(exec => ['wdiff', @opts, "$dir1/$cf", "$dir2/$cf"],
-		to_string => \$wdiff,
-		wait_child => 1);
-	};
-	if ($@ and $@ !~ /gave error exit status 1/) {
-	    print "$@\n";
-	    warn "wdiff failed\n";
-	} else {
-	    if (!$@) {
-		if (! $quiet) {
-		    print "\nNo differences were encountered between the $cf files$usepkgname\n";
-		}
-	    } elsif ($wdiff_opt) {
-		# Don't try messing with control codes
-		my $msg = ucfirst($cf) . " files$usepkgname: wdiff output";
-		print $msg, "\n", '-' x length $msg, "\n";
-		print $wdiff;
-	    } else {
-		my @output;
-		@output = split /\n/, $wdiff;
-		@output = grep /(\[-|\{\+)/, @output;
-		my $msg = ucfirst($cf) . " files$usepkgname: lines which differ (wdiff format)";
-		print "\n", $msg, "\n", '-' x length $msg, "\n";
-		print join("\n",@output), "\n";
-	    }
-	    $exit_status = 1;
-	}
-    }
     # Clean up
     rmtree([$dir1, $dir2]);
 }
@@ -1086,6 +991,77 @@ sub process_debI($)
     return \@filelist;
 }
 
+sub wdiff_control_files($$$$$)
+{
+    my ($dir1, $dir2, $debname, $controlfiles, $origstatus) = @_;
+    return unless defined $dir1 and defined $dir2 and defined $debname
+	and defined $controlfiles;
+    my @cf;
+    my $status = $origstatus;
+    if ($controlfiles eq 'ALL') {
+	# only need to list one directory as we are only comparing control
+	# files in both packages
+	@cf = grep { ! /md5sums/ } map { basename($_); } glob("$dir1/*");
+    } else {
+	@cf = split /,/, $controlfiles;
+    }
+
+    foreach my $cf (@cf) {
+	next unless -f "$dir1/$cf" and -f "$dir2/$cf";
+	if ($cf eq 'control' or $cf eq 'conffiles') {
+	    for my $file ("$dir1/$cf", "$dir2/$cf") {
+		my ($fd, @hdrs);
+		open $fd, '<', $file or fatal "Cannot read $file: $!";
+		while (<$fd>) {
+		    if (/^\s/ and @hdrs > 0) {
+			$hdrs[$#hdrs] .= $_;
+		    } else {
+			push @hdrs, $_;
+		    }
+		}
+		close $fd;
+		open $fd, '>', $file or fatal "Cannot write $file: $!";
+		print $fd sort @hdrs;
+		close $fd;
+	    }
+	}
+	my $usepkgname = $debname eq $dummyname ? "" : " of package $debname";
+	my @opts = ('-n');
+	push @opts, $wdiff_opt if $wdiff_opt;
+	my $wdiff = '';
+	eval {
+	    fork_and_exec(exec => ['wdiff', @opts, "$dir1/$cf", "$dir2/$cf"],
+		to_string => \$wdiff,
+		wait_child => 1);
+	};
+	if ($@ and $@ !~ /gave error exit status 1/) {
+	    print "$@\n";
+	    warn "wdiff failed\n";
+	} else {
+	    if (!$@) {
+		if (! $quiet) {
+		    print "\nNo differences were encountered between the $cf files$usepkgname\n";
+		}
+	    } elsif ($wdiff_opt) {
+		# Don't try messing with control codes
+		my $msg = ucfirst($cf) . " files$usepkgname: wdiff output";
+		print $msg, "\n", '-' x length $msg, "\n";
+		print $wdiff;
+	    } else {
+		my @output;
+		@output = split /\n/, $wdiff;
+		@output = grep /(\[-|\{\+)/, @output;
+		my $msg = ucfirst($cf) . " files$usepkgname: lines which differ (wdiff format)";
+		print "\n", $msg, "\n", '-' x length $msg, "\n";
+                print join("\n",@output), "\n";
+	    }
+	    $status = 1;
+	}
+    }
+
+    return $status;
+}
+
 sub mktmpdirs ()
 {
     no strict 'refs';
