@@ -41,8 +41,34 @@ PRECIOUS_FILES=0
 PROGNAME=`basename $0`
 MODIFIED_CONF_MSG='Default settings modified by devscripts configuration files:'
 
+# Temporary directories
+signingdir=""
+remotefilesdir=""
+
+trap "cleanup_tmpdir" 0 1 2 3 9 11 13 15
+
 # --- Functions
 
+mksigningdir () {
+    if [ -z "$signingdir" ]; then
+	signingdir="$(mktemp -dt debsign.XXXXXXXX)" || {
+	    echo "$PROGNAME: Can't create temporary directory" >&2
+	    echo "Aborting..." >&2
+	    exit 1
+	}
+    fi
+}
+
+mkremotefilesdir () {
+    if [ -z "$remotefilesdir" ]; then
+	remotefilesdir="$(mktemp -dt debsign.XXXXXXXX)" || {
+	    echo "$PROGNAME: Can't create temporary directory" >&2
+	    echo "Aborting..." >&2
+	    exit 1
+	}
+    fi
+}
+
 usage () {
     echo \
 "Usage: debsign [options] [changes, dsc or commands file]
@@ -62,6 +88,9 @@ usage () {
     -a<arch>        Use changes file made for Debian target architecture <arch>
     -t<target>      Use changes file made for GNU target architecture <target>
     --multi         Use most recent multiarch .changes file found
+    --debs-dir <directory>
+                    The location of the .changes / .dsc files when called from
+                    within a source tree (default "..")
     --no-conf, --noconf
                     Don't read devscripts config files;
                     must be the first option given
@@ -83,11 +112,44 @@ You are free to redistribute this code under the terms of the
 GNU General Public License, version 2 or later."
 }
 
-cleanup_tmpdir () {
-    if [ "$PRECIOUS_FILES" -gt 0 ]; then
-        echo "$PROGNAME: aborting with $PRECIOUS_FILES signed files in `pwd`" >&2
+temp_filename() {
+    local filename
+
+    if ! [ -w "$(dirname "$1")" ]; then
+	filename=`mktemp -t "$(basename "$1").$2.XXXXXXXXXX"` || {
+	    echo "$PROGNAME: Unable to create temporary file; aborting" >&2
+	    exit 1
+	}
     else
-        cd ..; rm -rf debsign.$$
+	filename="$1.$2"
+    fi
+
+    echo "$filename"
+}
+
+movefile() {
+    if [ -w "$(dirname "$2")" ]; then
+	mv -f -- "$1" "$2"
+    else
+	cat "$1" > "$2"
+	rm -f "$1"
+    fi
+}
+
+cleanup_tmpdir () {
+    if [ -n "$remotefilesdir" ] && [ -d "$remotefilesdir" ]; then
+	if [ "$PRECIOUS_FILES" -gt 0 ]; then
+	    echo "$PROGNAME: aborting with $PRECIOUS_FILES signed files in $remotefilesdir" >&2
+	    # Only produce the warning once...
+	    PRECIOUS_FILES=0
+	else
+	    cd ..
+	    rm -rf "$remotefilesdir"
+	fi
+    fi
+
+    if [ -n "$signingdir" ] && [ -d "$signingdir" ]; then
+	rm -rf "$signingdir"
     fi
 }
 
@@ -108,27 +170,33 @@ mustsetvar () {
 # information when this function is read first.
 signfile () {
     local savestty=$(stty -g 2>/dev/null) || true
+    mksigningdir
+    UNSIGNED_FILE="$signingdir/$(basename "$1")"
+    ASCII_SIGNED_FILE="${UNSIGNED_FILE}.asc"
+    (cat "$1" ; echo "") > "$UNSIGNED_FILE"
+
     if [ $signinterface = gpg ]
     then
 	gpgversion=`gpg --version | head -n 1 | cut -d' ' -f3`
 	gpgmajorversion=`echo $gpgversion | cut -d. -f1`
 	gpgminorversion=`echo $gpgversion | cut -d. -f2`
+
 	if [ $gpgmajorversion -gt 1 -o $gpgminorversion -ge 4 ]
 	then
-		(cat "$1" ; echo "") | \
-		    $signcommand --local-user "$2" --clearsign \
+		$signcommand --local-user "$2" --clearsign \
 		    --list-options no-show-policy-urls \
-		    --armor --textmode --output - - > "$1.asc" || \
+		    --armor --textmode --output "$ASCII_SIGNED_FILE"\
+		    "$UNSIGNED_FILE" || \
 		{ SAVESTAT=$?
 		  echo "$PROGNAME: gpg error occurred!  Aborting...." >&2
 		  stty $savestty 2>/dev/null || true
 		  exit $SAVESTAT
 		}
 	else
-		(cat "$1" ; echo "") | \
-		    $signcommand --local-user "$2" --clearsign \
-		        --no-show-policy-url \
-			--armor --textmode --output - - > "$1.asc" || \
+		$signcommand --local-user "$2" --clearsign \
+		    --no-show-policy-url \
+		    --armor --textmode --output "$ASCII_SIGNED_FILE" \
+		    "$UNSIGNED_FILE" || \
 		{ SAVESTAT=$?
 		  echo "$PROGNAME: gpg error occurred!  Aborting...." >&2
 		  stty $savestty 2>/dev/null || true
@@ -136,12 +204,13 @@ signfile () {
 		}
 	fi
     else
-	$signcommand -u "$2" +clearsig=on -fast < "$1" > "$1.asc"
+	$signcommand -u "$2" +clearsig=on -fast \
+	    < "$UNSIGNED_FILE" > "$ASCII_SIGNED_FILE"
     fi
     stty $savestty 2>/dev/null || true
     echo
     PRECIOUS_FILES=$(($PRECIOUS_FILES + 1))
-    mv -f -- "$1.asc" "$1"
+    movefile "$ASCII_SIGNED_FILE" "$1"
 }
 
 withecho () {
@@ -166,8 +235,10 @@ check_already_signed () {
 	fi
 	case $response in
 	[Nn]*)
-	    sed -e '1,/^$/d; /^$/,$d' "$1" > "$1.unsigned"
-	    mv "$1.unsigned" "$1"
+	    UNSIGNED_FILE="$(temp_filename "$1" "unsigned")"
+
+	    sed -e '1,/^$/d; /^$/,$d' "$1" > "$UNSIGNED_FILE"
+	    movefile "$UNSIGNED_FILE" "$1"
 	    return 1
 	    ;;
 	*) return 0;;
@@ -177,13 +248,18 @@ check_already_signed () {
 
 # --- main script
 
+# Unset GREP_OPTIONS for sanity
+unset GREP_OPTIONS
+
 # Boilerplate: set config variables
 DEFAULT_DEBSIGN_PROGRAM=
 DEFAULT_DEBSIGN_SIGNLIKE=
 DEFAULT_DEBSIGN_MAINT=
 DEFAULT_DEBSIGN_KEYID=
 DEFAULT_DEBSIGN_ALWAYS_RESIGN=
+DEFAULT_DEBRELEASE_DEBS_DIR=..
 VARS="DEBSIGN_PROGRAM DEBSIGN_SIGNLIKE DEBSIGN_MAINT DEBSIGN_KEYID DEBSIGN_ALWAYS_RESIGN"
+VARS="$VARS DEBRELEASE_DEBS_DIR"
 
 if [ "$1" = "--no-conf" -o "$1" = "--noconf" ]; then
     shift
@@ -208,7 +284,7 @@ else
 	  [ -r $file ] && . $file
 	done
 
-	set | egrep '^(DEBSIGN|DEVSCRIPTS)_')
+	set | egrep '^(DEBSIGN|DEBRELEASE|DEVSCRIPTS)_')
 
     # check sanity
     case "$DEBSIGN_SIGNLIKE" in
@@ -216,6 +292,13 @@ else
 	*) DEBSIGN_SIGNLIKE= ;;
     esac
 
+    # We do not replace this with a default directory to avoid accidentally
+    # signing a broken package
+    DEBRELEASE_DEBS_DIR="`echo \"$DEBRELEASE_DEBS_DIR\" | sed -e 's%/\+%/%g; s%\(.\)/$%\1%;'`"
+    if ! [ -d "$DEBRELEASE_DEBS_DIR" ]; then
+	debsdir_warning="config file specified DEBRELEASE_DEBS_DIR directory $DEBRELEASE_DEBS_DIR does not exist!"
+    fi
+
     # set config message
     MODIFIED_CONF=''
     for var in $VARS; do
@@ -234,6 +317,7 @@ fi
 
 maint="$DEBSIGN_MAINT"
 signkey="$DEBSIGN_KEYID"
+debsdir="$DEBRELEASE_DEBS_DIR"
 
 signcommand=''
 if [ -n "$DEBSIGN_PROGRAM" ]; then
@@ -279,6 +363,13 @@ do
 		    set -- "$@" "$filepart"
 		fi
 		;;
+	--debs-dir=*)
+	    opt_debsdir="`echo \"$1\" | sed -e 's/^--debs-dir=//; s%/\+%/%g; s%\(.\)/$%\1%;'`"
+	    ;;
+	--debs-dir)
+	    shift
+	    opt_debsdir="`echo \"$1\" | sed -e 's%/\+%/%g; s%\(.\)/$%\1%;'`"
+	    ;;
 	--no-conf|--noconf)
 		echo "$PROGNAME: $1 is only acceptable as the first command-line option!" >&2
 		exit 1 ;;
@@ -293,6 +384,18 @@ do
     shift
 done
 
+debsdir=${opt_debsdir:-$debsdir}
+# check sanity of debsdir
+if ! [ -d "$debsdir" ]; then
+    if [ -n "$debsdir_warning" ]; then
+        echo "$PROGNAME: $debsdir_warning" >&2
+        exit 1
+    else
+        echo "$PROGNAME: could not find directory $debsdir!" >&2
+        exit 1
+    fi
+fi
+
 if [ -z "$signcommand" ]; then
     echo "Could not find a signing program (pgp or gpg)!" >&2
     exit 1
@@ -313,10 +416,8 @@ dosigning() {
     # Do we have to download the changes file?
     if [ -n "$remotehost" ]
     then
-	cd ${TMPDIR:-/tmp}
-	mkdir debsign.$$ || { echo "$PROGNAME: Can't mkdir!" >&2; exit 1; }
-	trap "cleanup_tmpdir" 0 1 2 3 7 10 13 15
-	cd debsign.$$
+	mkremotefilesdir
+	cd "$remotefilesdir"
 
 	remotechanges=$changes
 	remotedsc=$dsc
@@ -327,11 +428,27 @@ dosigning() {
 	commands=`basename "$commands"`
 
 	if [ -n "$changes" ]
-	then withecho scp "$remotehost:$remotechanges" "$changes"
+	then
+	    if [ ! -f "$changes" ]
+	    then
+		withecho scp "$remotehost:$remotechanges" .
+	    fi
 	elif [ -n "$dsc" ]
 	then withecho scp "$remotehost:$remotedsc" "$dsc"
 	else withecho scp "$remotehost:$remotecommands" "$commands"
 	fi
+
+	if [ -n "$changes" ] && echo "$changes" | egrep -q '[][*?]'
+	then
+	    for changes in $changes
+	    do
+		printf "\n"
+		dsc=`echo "${remotedir+$remotedir/}$changes" | \
+		    perl -pe 's/\.changes$/.dsc/; s/(.*)_(.*)_(.*)\.dsc/\1_\2.dsc/'`
+		dosigning;
+	    done
+	    exit 0;
+	fi
     fi
 
     if [ -n "$changes" ]
@@ -344,7 +461,7 @@ dosigning() {
 
 	check_already_signed "$changes" "changes" && {
 	   echo "Leaving current signature unchanged." >&2
-	    exit 0
+	   return
 	}
 	if [ -n "$maint" ]
 	then maintainer="$maint"
@@ -375,10 +492,7 @@ dosigning() {
 	    dsc_sha1=`sha1sum $dsc | cut -d' ' -f1`
 	    dsc_sha256=`sha256sum $dsc | cut -d' ' -f1`
 
-	    temp_changes=`mktemp` || {
-		echo "$PROGNAME: Unable to create temporary changes file; aborting" >&2
-		exit 1
-	    }
+	    temp_changes="$(temp_filename "$changes" "temp")"
 	    cp "$changes" "$temp_changes"
 	    if perl -i -pe 'BEGIN {
 		'" \$dsc_file=\"$dsc\"; \$dsc_md5=\"$dsc_md5\"; "'
@@ -419,15 +533,15 @@ dosigning() {
 		    $insha256=0;
 		}' "$temp_changes"
 	    then
-		mv "$temp_changes" "$changes"
+		movefile "$temp_changes" "$changes"
 	    else
 		rm "$temp_changes"
 		echo "$PROGNAME: Error processing .changes file (see above)" >&2
 		exit 1
 	    fi
-	    
+
 	    withecho signfile "$changes" "$signas"
-	
+
 	    if [ -n "$remotehost" ]
 	    then
 		withecho scp "$changes" "$dsc" "$remotehost:$remotedir"
@@ -440,7 +554,7 @@ dosigning() {
 
 	    if [ -n "$remotehost" ]
 	    then
-		withecho scp "$changes" "$remotehost:$remotechanges"
+		withecho scp "$changes" "$remotehost:$remotedir"
 		PRECIOUS_FILES=$(($PRECIOUS_FILES - 1))
 	    fi
 
@@ -456,12 +570,11 @@ dosigning() {
 
 	check_already_signed "$commands" commands && {
 	    echo "Leaving current signature unchanged." >&2
-	    exit 0
+	    return
 	}
-    
-    
+
 	# simple validator for .commands files, see
-	# ftp://ftp-master.debian.org/pub/UploadQueue/README
+	# ftp://ftp.upload.debian.org/pub/UploadQueue/README
 	perl -ne 'BEGIN { $uploader = 0; $incommands = 0; }
               END { exit $? if $?;
                     if ($uploader && $incommands) { exit 0; }
@@ -469,8 +582,9 @@ dosigning() {
                   }
               sub checkcommands {
                   chomp($line=$_[0]);
-                  if ($line =~ m%^\s*mv(\s+[^\s/]+){2}\s*$%) { return 0; }
-                  if ($line =~ m%^\s*rm(\s+[^\s/]+)+\s*$%) { return 0; }
+                  if ($line =~ m%^\s*reschedule\s+[^\s/]+\.changes\s+[0-9]+-day\s*$%) { return 0; }
+                  if ($line =~ m%^\s*cancel\s+[^\s/]+\.changes\s*$%) { return 0; }
+                  if ($line =~ m%^\s*rm(\s+(?:DELAYED/[0-9]+-day/)?[^\s/]+)+\s*$%) { return 0; }
                   if ($line eq "") { return 0; }
                   die ".commands file has invalid Commands line: $line\n";
               }
@@ -488,13 +602,13 @@ dosigning() {
                  if (/./) { die ".commands file: extra stuff after Commands field!\n"; }
               }' $commands || {
 	echo "$PROGNAME: .commands file appears to be invalid. see:
-ftp://ftp-master.debian.org/pub/UploadQueue/README
+ftp://ftp.upload.debian.org/pub/UploadQueue/README
 for valid format" >&2;
 	exit 1; }
 
 	if [ -n "$maint" ]
 	then maintainer="$maint"
-	else 
+	else
             maintainer=`sed -n 's/^Uploader: //p' $commands`
             if [ -z "$maintainer" ]
             then
@@ -502,7 +616,7 @@ for valid format" >&2;
 		exit 1
             fi
 	fi
-    
+
 	signas="${signkey:-$maintainer}"
 
 	withecho signfile "$commands" "$signas"
@@ -523,7 +637,7 @@ for valid format" >&2;
 
 	check_already_signed "$dsc" dsc && {
 	    echo "Leaving current signature unchanged." >&2
-	    exit 0
+	    return
 	}
 	if [ -n "$maint" ]
 	then maintainer="$maint"
@@ -577,22 +691,22 @@ case $# in
 	sversion=`echo "$version" | perl -pe 's/^\d+://'`
 	pv="${package}_${sversion}"
 	pva="${package}_${sversion}_${arch}"
-	dsc="../$pv.dsc"
-	changes="../$pva.changes"
+	dsc="$debsdir/$pv.dsc"
+	changes="$debsdir/$pva.changes"
 	if [ -n "$multiarch" -o ! -r $changes ]; then
-	    changes=$(ls "../${package}_${sversion}_*+*.changes" "../${package}_${sversion}_multi.changes" 2>/dev/null | head -1)
+	    changes=$(ls "$debsdir/${package}_${sversion}_*+*.changes" "$debsdir/${package}_${sversion}_multi.changes" 2>/dev/null | head -1)
 	    if [ -z "$multiarch" ]; then
 		if [ -n "$changes" ]; then
 		    echo "$PROGNAME: could not find normal .changes file but found multiarch file:" >&2
 		    echo "  $changes" >&2
 		    echo "Using this changes file instead." >&2
-		else 
+		else
 		    echo "$PROGNAME: Can't find or can't read changes file $changes!" >&2
 		    exit 1
 		fi
 	    elif [ -n "$multiarch" -a -z "$changes" ]; then
 		echo "$PROGNAME: could not find any multiarch .changes file with name" >&2
-		echo "../${package}_${sversion}_*.changes" >&2
+		echo "$debsdir/${package}_${sversion}_*.changes" >&2
 		exit 1
 	    fi
 	fi
