@@ -56,11 +56,12 @@ if (system("command -v curl >/dev/null 2>&1") == 0) {
 
 sub usage {
     print <<"EOT";
-Usage: $progname [options] URL
+Usage: $progname [options] URL ...
        $progname [options] package[=version]
 
-Downloads Debian packages (source and binary) from the specified URL (first form),
+Downloads Debian packages (source and binary) from the specified URLs (first form),
 or using the mirror configured in /etc/apt/sources.list(.d) (second form).
+It is capable of downloading several packages at once.
 
    -b, --backup    Move files that would be overwritten to ./backup
    -q, --quiet     Suppress wget/curl output
@@ -104,7 +105,7 @@ sub wget {
     # schemes not supported by all backends
     if ($url =~ m!^(file|copy)://(/.+)!) {
 	if ($1 eq "copy" or not link($2, $file)) {
-	    system "cp -a $2 $file";
+	    system ('cp', '-a', $2, $file);
 	    return $? >> 8;
 	}
 	return;
@@ -116,7 +117,7 @@ sub wget {
     push @cmd, "-f", "-L" if $wget eq "curl";
     push @cmd, ($wget eq "wget" ? "-nv" : ("-s", "-S")) if $opt->{'quiet'};
     push @cmd, ($wget eq "wget" ? "--no-check-certificate" : "--insecure") if $opt->{'insecure'};
-    push @cmd, ($wget eq "wget" ? "--no-chache" : ("--header", "Pragma: no-cache")) if $opt->{'no-cache'};
+    push @cmd, ($wget eq "wget" ? "--no-cache" : ("--header", "Pragma: no-cache")) if $opt->{'no-cache'};
     push @cmd, ($wget eq "wget" ? "-O" : "-o");
     system @cmd, $file, $url;
     return $? >> 8;
@@ -152,9 +153,9 @@ sub get_file {
 	my $md5sum_new = Digest::MD5->new->addfile($fh5)->hexdigest();
 	close $fh5;
 	if (not $md5sum or ($md5sum_new eq $md5sum)) {
-	    print "$progname: using existing $file\n";
+	    print "$progname: using existing $file\n" unless $opt->{'quiet'};
 	} else {
-	    print "$progname: removing $file (md5sum does not match)\n";
+	    print "$progname: removing $file (md5sum does not match)\n" unless $opt->{'quiet'};
 	    backup_or_unlink($file);
 	}
     }
@@ -171,10 +172,10 @@ sub get_file {
 
 	    if ($md5sum_new eq $md5sum) {
 		if (link "$path/$file", $file) {
-		    print "$progname: using $path/$file (hardlink)\n";
+		    print "$progname: using $path/$file (hardlink)\n" unless $opt->{'quiet'};
 		} else {
-		    print "$progname: using $path/$file (copy)\n";
-		    system "cp -a $path/$file $file";
+		    print "$progname: using $path/$file (copy)\n" unless $opt->{'quiet'};
+		    system ('cp', '-a', "$path/$file", $file);
 		}
 		last;
 	    }
@@ -183,7 +184,7 @@ sub get_file {
 
     # finally get it from the web
     unless (-e $file) {
-	print "$progname: retrieving $dir/$file\n";
+	print "$progname: retrieving $dir/$file\n" unless $opt->{'quiet'};
 	if (wget($file, "$dir/$file")) {
 	    warn "$progname: $wget $file $dir/$file failed\n";
 	    unlink $file;
@@ -191,7 +192,7 @@ sub get_file {
     }
 
     # try apt-get if it is still not there
-    if (not -e $file and $file =~ m!^([a-z0-9.+-]{2,})_[^/]+\.(?:diff\.gz|tar\.gz)$!) {
+    if (not -e $file and $file =~ m!^([a-z0-9][a-z0-9.+-]+)_[^/]+\.(?:diff\.gz|tar\.gz)$!) {
 	my $cmd = "apt-get source --print-uris $1";
 	my $apt = new IO::File("$cmd |") or die "$cmd: $!";
 	while(<$apt>) {
@@ -228,7 +229,10 @@ sub parse_file {
     open $fh, $file or die "$file: $!";
     while (<$fh>) {
 	if (/^ ([0-9a-f]{32}) (?:\S+ )*(\S+)$/) {
-	    get_file($dir, $2, $1) or return;
+	    my ($_sum, $_file) = ($1, $2);
+	    $_file !~ m,[/\x00],
+		or die "File name contains invalid characters: $_file";
+            get_file($dir, $_file, $_sum) or return;
 	}
     }
     close $fh;
@@ -438,21 +442,23 @@ for my $arg (@ARGV) {
 	    if ($dget_verify) { # We are duplicating work here a bit as
 		# dpkg-source -x will also verify signatures. Still, we
 		# also want to barf with -d, and on unsigned packages.
-		system 'dscverify', $found_dsc;
+		system ('dscverify', $found_dsc);
 		exit $? >> 8 if $? >> 8 != 0;
 	    }
+	    my @cmd = qw(dpkg-source -x);
+	    push @cmd, '--no-check' unless $dget_verify;
 	    if ($opt->{'build'}) {
-		my @output = `dpkg-source -x $found_dsc`; # FIXME: this will break when dpkg-source output is localized
-		print @output;
+		my @output = `LC_ALL=C @cmd $found_dsc`;
+		print @output unless $opt->{'quiet'};
 		foreach (@output) {
-		    if ( /^dpkg-source: extracting .* in (.*)/ ) {
+		    if ( /^dpkg-source: (?:info: )?extracting .* in (.*)/ ) {
 			chdir $1;
 			system 'dpkg-buildpackage', '-b', '-uc';
 			last;
 		    }
 		}
 	    } elsif ($dget_unpack) {
-		system 'dpkg-source', '-x', $found_dsc;
+		system @cmd, $found_dsc;
 	    }
 	}
 
@@ -479,7 +485,7 @@ dget -- Download Debian source and binary packages
 
 =over
 
-=item B<dget> [I<options>] I<URL>
+=item B<dget> [I<options>] I<URL> ...
 
 =item B<dget> [I<options>] I<package>[=I<version>]
 
@@ -488,7 +494,7 @@ dget -- Download Debian source and binary packages
 =head1 DESCRIPTION
 
 B<dget> downloads Debian packages.  In the first form, B<dget> fetches
-the requested URL.  If this is a .dsc or .changes file, then B<dget>
+the requested URLs.  If this is a .dsc or .changes file, then B<dget>
 acts as a source-package aware form of B<wget>: it also fetches any
 files referenced in the .dsc/.changes file.  The downloaded source is
 then unpacked by B<dpkg-source>.
@@ -500,6 +506,14 @@ require root privileges, writes to the current directory, and does not
 download dependencies.  If a version number is specified, this version
 of the package is requested.
 
+In both cases dget is capable of getting several packages and/or URLs
+at once.
+
+(Note that I<.udeb> packages used by debian-installer are located in separate
+packages files from I<.deb> packages. In order to use I<.udebs> with B<dget>,
+you will need to have configured B<apt> to use a packages file for
+I<component>/I<debian-installer>).
+
 Before downloading files listed in .dsc and .changes files, and before
 downloading binary packages, B<dget> checks to see whether any of
 these files already exist.  If they do, then their md5sums are
@@ -605,7 +619,7 @@ is 'yes'.
 Set to 'no' to disable checking signatures of downloaded source
 packages.  Default is 'yes'.
 
-=cut
+=back
 
 =head1 BUGS AND COMPATIBILITY
 
