@@ -31,9 +31,14 @@ usage () {
     --sendmail=SENDMAILCMD
                       Use SENDMAILCMD instead of \"/usr/sbin/sendmail -t\"
     --mutt            Use mutt to mail the message (default)
-    --no-mutt         Mail the message directly, don\'t use mutt
+    --no-mutt         Mail the message directly, don't use mutt
     --from=EMAIL      Use EMAIL address for message to BTS; defaults to
                       value of DEBEMAIL or EMAIL
+    --delay=DELAY     Indicate that the NMU has been uploaded to the
+                      DELAYED queue, with a delay of DELAY days; defaults
+                      to a placeholder value of "XX". If DELAY is 0 then
+                      no reference is made to the DELAYED queue.
+    --no-delay        Equivalent to \"--delay=0\"
     --no-conf, --noconf
                       Don't read devscripts config files;
                       must be the first option given
@@ -54,13 +59,14 @@ EOF
 }
 
 # Boilerplate: set config variables
+DEFAULT_NMUDIFF_DELAY="XX"
 DEFAULT_NMUDIFF_MUTT="yes"
 DEFAULT_NMUDIFF_NEWREPORT="maybe"
 DEFAULT_BTS_SENDMAIL_COMMAND="/usr/sbin/sendmail"
-VARS="NMUDIFF_MUTT NMUDIFF_NEWREPORT BTS_SENDMAIL_COMMAND"
+VARS="NMUDIFF_DELAY NMUDIFF_MUTT NMUDIFF_NEWREPORT BTS_SENDMAIL_COMMAND"
 # Don't think it's worth including this stuff
 # DEFAULT_DEVSCRIPTS_CHECK_DIRNAME_LEVEL=1
-# DEFAULT_DEVSCRIPTS_CHECK_DIRNAME_REGEX='PACKAGE(-.*)?'
+# DEFAULT_DEVSCRIPTS_CHECK_DIRNAME_REGEX='PACKAGE(-.+)?'
 # VARS="BTS_SENDMAIL_COMMAND DEVSCRIPTS_CHECK_DIRNAME_LEVEL DEVSCRIPTS_CHECK_DIRNAME_REGEX"
 
 if [ "$1" = "--no-conf" -o "$1" = "--noconf" ]; then
@@ -93,6 +99,14 @@ else
 	"") BTS_SENDMAIL_COMMAND=/usr/sbin/sendmail ;;
 	*) ;;
     esac
+    if [ "$NMUDIFF_DELAY" = "XX" ]; then
+	# Fine
+	:
+    else
+	if ! [ "$NMUDIFF_DELAY" -ge 0 ] 2>/dev/null; then
+	    NMUDIFF_DELAY=XX
+	fi
+    fi
     case "$NMUDIFF_MUTT" in
 	yes|no) ;;
 	*) NMUDIFF_MUTT=yes ;;
@@ -130,6 +144,7 @@ fi
 # Removed: --long check-dirname-level:,check-dirname-regex: \
 TEMP=$(getopt -s bash -o "h" \
 	--long sendmail:,from:,new,old,mutt,no-mutt,nomutt \
+	--long delay:,no-delay,nodelay \
 	--long no-conf,noconf \
 	--long help,version -n "$PROGNAME" -- "$@")
 if [ $? != 0 ] ; then exit 1 ; fi
@@ -149,6 +164,21 @@ while [ "$1" ]; do
 # 	;;
 #     --check-dirname-regex)
 # 	shift; 	CHECK_DIRNAME_REGEX="$1" ;;
+    --delay)
+	shift
+	if [ "$1" = "XX" ]; then
+	    # Fine
+	    NMUDIFF_DELAY="$1"
+	else
+	    if ! [ "$1" -ge 0 ] 2>/dev/null; then
+		NMUDIFF_DELAY=XX
+	    else
+		NMUDIFF_DELAY="$1"
+	    fi
+	fi
+	;;
+    --nodelay|--no-delay)
+	NMUDIFF_DELAY=0 ;;
     --mutt)
 	NMUDIFF_MUTT=yes ;;
     --nomutt|--no-mutt)
@@ -193,6 +223,10 @@ if [ "$NMUDIFF_MUTT" = yes ] && ! command -v mutt > /dev/null 2>&1; then
 fi
 
 if [ "$NMUDIFF_MUTT" = no ]; then
+    if [ -z "$FROM" ]; then
+	: ${FROMNAME:="$DEBFULLNAME"}
+	: ${FROMNAME:="$NAME"}
+    fi
     : ${FROM:="$DEBEMAIL"}
     : ${FROM:="$EMAIL"}
     if [ -z "$FROM" ]; then
@@ -200,6 +234,13 @@ if [ "$NMUDIFF_MUTT" = no ]; then
 	echo "or EMAIL environment variable or using --from command line option." >&2
 	exit 1
     fi
+    if [ -n "$FROMNAME" ]; then
+	# If $FROM looks like "Name <email@address>" then extract just the address
+	if [ "$FROM" = "$(echo "$FROM" | sed -ne '/^\(.*\) *<\(.*\)> *$/p')" ]; then
+	    FROM="$(echo "$FROM" | sed -ne 's/^\(.*\) *<\(.*\)> *$/\2/p')"
+	fi
+	FROM="$FROMNAME <$FROM>"
+    fi
 fi
 
 if ! [ -f debian/changelog ]; then
@@ -254,10 +295,11 @@ if [ ! -r ../${SOURCE}_${VERSION_NO_EPOCH}.dsc ]; then
     exit 1
 fi
 
+ret=0
 debdiff ../${SOURCE}_${OLDVERSION_NO_EPOCH}.dsc \
   ../${SOURCE}_${VERSION_NO_EPOCH}.dsc \
-  > ../${SOURCE}-${VERSION_NO_EPOCH}-nmu.diff
-if [ $? -ne 0 ]; then
+  > ../${SOURCE}-${VERSION_NO_EPOCH}-nmu.diff || ret=$?
+if [ $ret -ne 0 ] && [ $ret -ne 1 ]; then
     echo "nmudiff: debdiff failed, aborting." >&2
     rm -f ../${SOURCE}-${VERSION_NO_EPOCH}-nmu.diff
     exit 1
@@ -282,8 +324,10 @@ else
 	TO_ADDRESSES_SENDMAIL="$TO_ADDRESSES_SENDMAIL,
   $b@bugs.debian.org"
 	TO_ADDRESSES_MUTT="$TO_ADDRESSES_MUTT $b@bugs.debian.org"
-	TAGS="$TAGS
+	if [ "`bts select bugs:$b tag:patch`" != "$b" ]; then
+	    TAGS="$TAGS
 tags $b + patch"
+	fi
     done
     TO_ADDRESSES_SENDMAIL=$(echo "$TO_ADDRESSES_SENDMAIL" | tail -n +2)
     TAGS=$(echo "$TAGS" | tail -n +2)
@@ -293,6 +337,22 @@ fi
 
 TMPNAM="$( tempfile )"
 
+if [ "$NMUDIFF_DELAY" = "XX" ]; then
+    DELAY_HEADER="
+[Replace XX with correct value]"
+fi
+
+if [ "$NMUDIFF_DELAY" = "0" ]; then
+    BODY="$(printf "%s\n%s\n" \
+"I've prepared an NMU for $SOURCE (versioned as $VERSION). The diff" \
+"is attached to this message.")"
+else
+    BODY="$(printf "%s\n%s\n%s\n" \
+"I've prepared an NMU for $SOURCE (versioned as $VERSION) and" \
+"uploaded it to DELAYED/$NMUDIFF_DELAY. Please feel free to tell me if I" \
+"should delay it longer.")"
+fi
+
 if [ "$NMUDIFF_MUTT" = no ]; then
     cat <<EOF > "$TMPNAM"
 From: $FROM
@@ -304,11 +364,12 @@ Date: `date -R`
 X-NMUDIFF-Version: ###VERSION###
 
 $TAGS
+$DELAY_HEADER
+Dear maintainer,
 
-Hi,
-
-The following is the diff for my $SOURCE $VERSION NMU.
+$BODY
 
+Regards.
 EOF
 
     cat ../${SOURCE}-${VERSION_NO_EPOCH}-nmu.diff >> "$TMPNAM"
@@ -319,15 +380,17 @@ EOF
 	exit 1
     fi
 
-    echo -n "Do you want to go ahead and submit the bug report now? (y/n) "
-    read response
-    case "$response" in
-	y*) ;;
-	*)  echo "OK, then, aborting." >&2
-	    rm -f ../${SOURCE}-${VERSION_NO_EPOCH}-nmu.diff "$TMPNAM"
-	    exit 1
-	    ;;
-    esac
+    while : ; do
+	echo -n "Do you want to go ahead and submit the bug report now? (y/n) "
+	read response
+	case "$response" in
+	    y*) break;;
+	    n*) echo "OK, then, aborting." >&2
+		rm -f ../${SOURCE}-${VERSION_NO_EPOCH}-nmu.diff "$TMPNAM"
+		exit 1
+		;;
+	esac
+    done
 
     case "$BTS_SENDMAIL_COMMAND" in
 	/usr/sbin/sendmail*|/usr/sbin/exim*)
@@ -340,16 +403,18 @@ EOF
 else # NMUDIFF_MUTT=yes
     cat <<EOF > "$TMPNAM"
 $TAGS
+$DELAY_HEADER
+Dear maintainer,
 
-Hi,
+$BODY
 
-Attached is the diff for my $SOURCE $VERSION NMU.
+Regards.
 EOF
 
     mutt -s "$SOURCE: diff for NMU version $VERSION" -i "$TMPNAM" \
-	-a ../${SOURCE}-${VERSION_NO_EPOCH}-nmu.diff \
 	-e "my_hdr X-NMUDIFF-Version: ###VERSION###" \
-	$BCC_ADDRESS_MUTT $TO_ADDRESSES_MUTT
+	-a ../${SOURCE}-${VERSION_NO_EPOCH}-nmu.diff $BCC_ADDRESS_MUTT \
+	-- $TO_ADDRESSES_MUTT
 
 fi
 
