@@ -25,9 +25,11 @@
 use 5.008;  # uses 'our' variables and filetest
 use strict;
 use Cwd;
+use Cwd 'abs_path';
+use Dpkg::IPC;
 use File::Basename;
 use File::Copy;
-use File::Temp qw/tempdir/;
+use File::Temp qw/tempfile tempdir/;
 use filetest 'access';
 use Getopt::Long;
 use lib '/usr/share/devscripts';
@@ -108,7 +110,7 @@ Options:
                    What constitutes a matching directory name; REGEX is
                    a Perl regular expression; the string \`PACKAGE\' will
                    be replaced by the package name; see manpage for details
-                   (default: 'PACKAGE(-.*)?')
+                   (default: 'PACKAGE(-.+)?')
     --watchfile FILE
                    Specify the watchfile rather than using debian/watch;
                    no directory traversing will be done in this case
@@ -119,6 +121,8 @@ Options:
                    Specify the version which the upstream release must
                    match in order to be considered, rather than using the
                    release with the highest version
+    --download-current-version
+                   Download the currently packaged version
     --package PACKAGE
                    Specify the package name rather than examining
                    debian/changelog; must use --upstream-version and
@@ -166,7 +170,7 @@ my $repack = 0; # repack .tar.bz2 or .zip to .tar.gz
 my $symlink = 'symlink';
 my $verbose = 0;
 my $check_dirname_level = 1;
-my $check_dirname_regex = 'PACKAGE(-.*)?';
+my $check_dirname_regex = 'PACKAGE(-.+)?';
 my $dehs = 0;
 my %dehs_tags;
 my $dehs_end_output = 0;
@@ -189,8 +193,9 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 		       'USCAN_VERBOSE' => 'no',
 		       'USCAN_DEHS_OUTPUT' => 'no',
 		       'USCAN_USER_AGENT' => '',
+		       'USCAN_REPACK' => 'no',
 		       'DEVSCRIPTS_CHECK_DIRNAME_LEVEL' => 1,
-		       'DEVSCRIPTS_CHECK_DIRNAME_REGEX' => 'PACKAGE(-.*)?',
+		       'DEVSCRIPTS_CHECK_DIRNAME_REGEX' => 'PACKAGE(-.+)?',
 		       );
     my %config_default = %config_vars;
 
@@ -224,6 +229,8 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 	or $config_vars{'USCAN_VERBOSE'}='no';
     $config_vars{'USCAN_DEHS_OUTPUT'} =~ /^(yes|no)$/
 	or $config_vars{'USCAN_DEHS_OUTPUT'}='no';
+    $config_vars{'USCAN_REPACK'} =~ /^(yes|no)$/
+	or $config_vars{'USCAN_REPACK'}='no';
     $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_LEVEL'} =~ /^[012]$/
 	or $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_LEVEL'}=1;
 
@@ -248,6 +255,7 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     $check_dirname_regex = $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_REGEX'};
     $user_agent_string = $config_vars{'USCAN_USER_AGENT'}
 	if $config_vars{'USCAN_USER_AGENT'};
+    $repack = $config_vars{'USCAN_REPACK'} eq 'yes' ? 1 : 0;
 }
 
 # Now read the command line arguments
@@ -258,6 +266,7 @@ my ($opt_verbose, $opt_ignore, $opt_level, $opt_regex, $opt_noconf);
 my ($opt_package, $opt_uversion, $opt_watchfile, $opt_dehs, $opt_timeout);
 my $opt_download_version;
 my $opt_user_agent;
+my $opt_download_current_version;
 
 GetOptions("help" => \$opt_h,
 	   "version" => \$opt_v,
@@ -285,6 +294,7 @@ GetOptions("help" => \$opt_h,
 	   "useragent=s" => \$opt_user_agent,
 	   "noconf" => \$opt_noconf,
 	   "no-conf" => \$opt_noconf,
+	   "download-current-version" => \$opt_download_current_version,
 	   )
     or die "Usage: $progname [options] [directories]\nRun $progname --help for more details\n";
 
@@ -362,7 +372,30 @@ else { $passive = undef; }
 #   else { delete $ENV{'FTP_PASSIVE'}; }
 # to restore $ENV{'FTP_PASSIVE'} to what it was at this point
 
-my $user_agent = LWP::UserAgent->new(env_proxy => 1);
+# dummy subclass used to store all the redirections for later use
+package LWP::UserAgent::UscanCatchRedirections;
+
+use base 'LWP::UserAgent';
+
+my @uscan_redirections;
+
+sub redirect_ok {
+    my $self = shift;
+    my ($request) = @_;
+    if ($self->SUPER::redirect_ok(@_)) {
+	push @uscan_redirections, $request->uri;
+	return 1;
+    }
+    return 0;
+}
+
+sub get_redirections {
+    return \@uscan_redirections;
+}
+
+package main;
+
+my $user_agent = LWP::UserAgent::UscanCatchRedirections->new(env_proxy => 1);
 $user_agent->timeout($timeout);
 $user_agent->agent($user_agent_string);
 
@@ -427,7 +460,7 @@ if (defined $opt_watchfile) {
     # Are there any warnings to give if we're using dehs?
     $dehs_end_output=1;
     dehs_output if $dehs;
-    exit 0;
+    exit ($found ? 0 : 1);
 }
 
 # Otherwise we're scanning for watchfiles
@@ -656,13 +689,19 @@ sub process_watchline ($$$$$$)
 
     my $origline = $line;
     my ($base, $site, $dir, $filepattern, $pattern, $lastversion, $action);
+    my $basedir;
+    my (@patterns, @sites, @redirections, @basedirs);
     my %options = ();
 
     my ($request, $response);
     my ($newfile, $newversion);
     my $style='new';
     my $urlbase;
+    my $headers = HTTP::Headers->new;
 
+    # Comma-separated list of features that sites being queried might
+    # want to be aware of
+    $headers->header('X-uscan-features' => 'enhanced-matching');
     %dehs_tags = ('package' => $pkg);
 
     if ($watch_version == 1) {
@@ -764,7 +803,10 @@ sub process_watchline ($$$$$$)
 	}
 
 	# Handle sf.net addresses specially
-	$base =~ s%^http://sf\.net/%http://qa.debian.org/watch/sf.php/%;
+	if ($base =~ m%^http://sf\.net/%) {
+	    $base =~ s%^http://sf\.net/%http://qa.debian.org/watch/sf.php/%;
+	    $filepattern .= '(?:\?.*)?';
+	}
 	if ($base =~ m%^(\w+://[^/]+)%) {
 	    $site = $1;
 	} else {
@@ -779,17 +821,47 @@ sub process_watchline ($$$$$$)
 	# We're going to make the pattern
 	# (?:(?:http://site.name)?/dir/path/)?base_pattern
 	# It's fine even for ftp sites
-	my $basedir = $base;
+	$basedir = $base;
 	$basedir =~ s%^\w+://[^/]+/%/%;
 	$pattern = "(?:(?:$site)?" . quotemeta($basedir) . ")?$filepattern";
     }
 
+    if (! $lastversion or $lastversion eq 'debian') {
+	if (defined $pkg_version) {
+	    $lastversion=$pkg_version;
+	} else {
+	    warn "$progname warning: Unable to determine current version\n  in $watchfile, skipping:\n  $line\n";
+	    return 1;
+	}
+    }
+    # And mangle it if requested
+    my $mangled_lastversion;
+    $mangled_lastversion = $lastversion;
+    foreach my $pat (@{$options{'dversionmangle'}}) {
+	if (! safe_replace(\$mangled_lastversion, $pat)) {
+	    warn "$progname: In $watchfile, potentially"
+	      . " unsafe or malformed dversionmangle"
+	      . " pattern:\n  '$pat'"
+	      . " found. Skipping watchline\n"
+	      . "  $line\n";
+	    return 1;
+	}
+    }
+    if($opt_download_current_version) {
+	$download_version = $mangled_lastversion;
+	$force_download = 1;
+    }
+
     # Check all's OK
     if ($pattern !~ /\(.*\)/) {
 	warn "$progname warning: Filename pattern missing version delimiters ()\n  in $watchfile, skipping:\n  $line\n";
 	return 1;
     }
 
+    push @patterns, $pattern;
+    push @sites, $site;
+    push @basedirs, $basedir;
+
     # What is the most recent file, based on the filenames?
     # We first have to find the candidates, then we sort them using
     # Devscripts::Versort::versort
@@ -798,13 +870,40 @@ sub process_watchline ($$$$$$)
 	    die "$progname: you must have the libcrypt-ssleay-perl package installed\nto use https URLs\n";
 	}
 	print STDERR "$progname debug: requesting URL $base\n" if $debug;
-	$request = HTTP::Request->new('GET', $base);
+	$request = HTTP::Request->new('GET', $base, $headers);
 	$response = $user_agent->request($request);
 	if (! $response->is_success) {
 	    warn "$progname warning: In watchfile $watchfile, reading webpage\n  $base failed: " . $response->status_line . "\n";
 	    return 1;
 	}
 
+	@redirections = @{$user_agent->get_redirections};
+	
+	print STDERR "$progname debug: redirections: @redirections\n"
+	    if $debug;
+
+	foreach my $_redir (@redirections) {
+	    my $base_dir = $_redir;
+	    
+	    $base_dir =~ s%^\w+://[^/]+/%/%;
+	    if ($_redir =~ m%^(\w+://[^/]+)%) {
+		my $base_site = $1;
+
+		push @patterns, "(?:(?:$base_site)?" . quotemeta($base_dir) . ")?$filepattern";
+		push @sites, $base_site;
+		push @basedirs, $base_dir;
+
+		# remove the filename, if any
+		my $base_dir_orig = $base_dir;
+		$base_dir =~ s%/[^/]*$%/%;
+		if ($base_dir ne $base_dir_orig) {
+		    push @patterns, "(?:(?:$base_site)?" . quotemeta($base_dir) . ")?$filepattern";
+		    push @sites, $base_site;
+		    push @basedirs, $base_dir;
+		}
+	    }
+	}
+
 	my $content = $response->content;
 	print STDERR "$progname debug: received content:\n$content\[End of received content]\n"
 	    if $debug;
@@ -815,7 +914,7 @@ sub process_watchline ($$$$$$)
 	# Strip comments
 	$content =~ s/<!-- .*?-->//sg;
 	# Is there a base URL given?
-	if ($content =~ /<\s*base\s+[^>]*href\s*=\s*([\"\'])(.*?)\1/) {
+	if ($content =~ /<\s*base\s+[^>]*href\s*=\s*([\"\'])(.*?)\1/i) {
 	    # Ensure it ends with /
 	    $urlbase = "$2/";
 	    $urlbase =~ s%//$%/%;
@@ -824,33 +923,36 @@ sub process_watchline ($$$$$$)
 	    ($urlbase = $base) =~ s%/[^/]*$%/%;
 	}
 
-	print STDERR "$progname debug: matching pattern $pattern\n" if $debug;
+	print STDERR "$progname debug: matching pattern(s) @patterns\n" if $debug;
 	my @hrefs;
-	while ($content =~ m/<\s*a\s+[^>]*href\s*=\s*([\"\'])(.*?)\1/gi) {
+	while ($content =~ m/<\s*a\s+[^>]*href\s*=\s*([\"\'])(.*?)\1/sgi) {
 	    my $href = $2;
-	    if ($href =~ m&^$pattern$&) {
-		if ($watch_version == 2) {
-		    # watch_version 2 only recognised one group; the code
-		    # below will break version 2 watchfiles with a construction
-		    # such as file-([\d\.]+(-\d+)?) (bug #327258)
-		    push @hrefs, [$1, $href];
-		} else {
-		    # need the map { ... } here to handle cases of (...)?
-		    # which may match but then return undef values
-		    my $mangled_version =
-			join(".", map { $_ if defined($_) }
-			     $href =~ m&^$pattern$&);
-		    foreach my $pat (@{$options{'uversionmangle'}}) {
-                       if (! safe_replace(\$mangled_version, $pat)) {
-                           warn "$progname: In $watchfile, potentially"
-                             . " unsafe or malformed uversionmangle"
-                             . " pattern:\n  '$pat'"
-                             . " found. Skipping watchline\n"
-                             . "  $line\n";
-                           return 1;
-                       }
+	    $href =~ s/\n//g;
+	    foreach my $_pattern (@patterns) {
+		if ($href =~ m&^$_pattern$&) {
+		    if ($watch_version == 2) {
+			# watch_version 2 only recognised one group; the code
+			# below will break version 2 watchfiles with a construction
+			# such as file-([\d\.]+(-\d+)?) (bug #327258)
+			push @hrefs, [$1, $href];
+		    } else {
+			# need the map { ... } here to handle cases of (...)?
+			# which may match but then return undef values
+			my $mangled_version =
+			    join(".", map { $_ if defined($_) }
+			 	$href =~ m&^$_pattern$&);
+			foreach my $pat (@{$options{'uversionmangle'}}) {
+			    if (! safe_replace(\$mangled_version, $pat)) {
+				warn "$progname: In $watchfile, potentially"
+			 	 . " unsafe or malformed uversionmangle"
+				  . " pattern:\n  '$pat'"
+				  . " found. Skipping watchline\n"
+				  . "  $line\n";
+				return 1;
+			    }
+			}
+			push @hrefs, [$mangled_version, $href];
 		    }
-		    push @hrefs, [$mangled_version, $href];
 		}
 	    }
 	}
@@ -914,7 +1016,7 @@ sub process_watchline ($$$$$$)
 	# that we can target our search correctly
 	if ($content =~ /<\s*a\s+[^>]*href/i) {
 	    while ($content =~ 
-		m/(?:<\s*a\s+[^>]*href\s*=\s*\")($pattern)\"/gi) {
+		m/(?:<\s*a\s+[^>]*href\s*=\s*\")((?-i)$pattern)\"/gi) {
 		my $file = $1;
 		my $mangled_version = join(".", $file =~ m/^$pattern$/);
 		foreach my $pat (@{$options{'uversionmangle'}}) {
@@ -932,7 +1034,7 @@ sub process_watchline ($$$$$$)
 	} else {
 	    # they all look like:
 	    # info info ... info filename [ -> linkname]
-	    while ($content =~ m/($filepattern)(\s+->\s+\S+)?$/mgi) {
+	    while ($content =~ m/\s($filepattern)(\s+->\s+\S+)?$/mg) {
 		my $file = $1;
 		my $mangled_version = join(".", $file =~ m/^$filepattern$/);
 		foreach my $pat (@{$options{'uversionmangle'}}) {
@@ -1016,27 +1118,7 @@ EOF
 	    $newfile_base = "$pkg-$newversion.download";
 	}
     }
-    if (! $lastversion or $lastversion eq 'debian') {
-	if (defined $pkg_version) {
-	    $lastversion=$pkg_version;
-	} else {
-	    warn "$progname warning: Unable to determine current version\n  in $watchfile, skipping:\n  $line\n";
-	    return 1;
-	}
-    }
-    # And mangle it if requested
-    my $mangled_lastversion = $lastversion;
-    foreach my $pat (@{$options{'dversionmangle'}}) {
-	if (! safe_replace(\$mangled_lastversion, $pat)) {
-	    warn "$progname: In $watchfile, potentially"
-	      . " unsafe or malformed dversionmangle"
-	      . " pattern:\n  '$pat'"
-	      . " found. Skipping watchline\n"
-	      . "  $line\n";
-	    return 1;
-	}
-    }
-
+    
     # So what have we got to report now?
     my $upstream_url;
     # Upstream URL?  Copying code from below - ugh.
@@ -1047,11 +1129,50 @@ EOF
 	}
 	# absolute filename?
 	elsif ($newfile =~ m%^/%) {
-	    $upstream_url = "$site$newfile";
+	    # Were there any redirections? If so try using those first
+	    if ($#patterns > 0) {
+		# replace $site here with the one we were redirected to
+		foreach my $index (0 .. $#patterns) {
+		    if ("$sites[$index]$newfile" =~ m&^$patterns[$index]$&) {
+			$upstream_url = "$sites[$index]$newfile";
+			last;
+		    }
+		}
+		if (!defined($upstream_url)) {
+		    if ($debug) {
+			warn "$progname warning: Unable to determine upstream url from redirections,\n" .
+			    "defaulting to using site specified in watchfile\n";
+		    }
+		    $upstream_url = "$sites[0]$newfile";
+		}
+	    } else {
+		$upstream_url = "$sites[0]$newfile";
+	    }
 	}
 	# relative filename, we hope
 	else {
-	    $upstream_url = "$urlbase$newfile";
+	    # Were there any redirections? If so try using those first
+	    if ($#patterns > 0) {
+		# replace $site here with the one we were redirected to
+		foreach my $index (0 .. $#patterns) {
+		    # skip unless the basedir looks like a directory
+		    next unless $basedirs[$index] =~ m%/$%;
+		    my $nf = "$basedirs[$index]$newfile";
+		    if ("$sites[$index]$nf" =~ m&^$patterns[$index]$&) {
+			$upstream_url = "$sites[$index]$nf";
+			last;
+		    }
+		}
+		if (!defined($upstream_url)) {
+		    if ($debug) {
+			warn "$progname warning: Unable to determine upstream url from redirections,\n" .
+			    "defaulting to using site specified in watchfile\n";
+		    }
+		    $upstream_url = "$urlbase$newfile";
+		}
+	    } else {
+		$upstream_url = "$urlbase$newfile";
+	    }
 	}
 
 	# mangle if necessary
@@ -1108,7 +1229,7 @@ EOF
     # We use dpkg's rules to determine whether our current version
     # is newer or older than the remote version.
     if (!defined $download_version) {
-        if (system("dpkg", "--compare-versions", "'$mangled_lastversion'", "gt", "'$newversion'") == 0) {
+	if (system("dpkg", "--compare-versions", "'$mangled_lastversion'", "gt", "'$newversion'") == 0) {
 	    if ($verbose) {
 		print " => remote site does not even have current version\n";
 	    } elsif ($dehs) {
@@ -1225,8 +1346,29 @@ EOF
     if ($repack and $newfile_base =~ /^(.*)\.(tar\.bz2|tbz2?)$/) {
 	print "-- Repacking from bzip2 to gzip\n" if $verbose;
 	my $newfile_base_gz = "$1.tar.gz";
-	system("bunzip2 -c $destdir/$newfile_base | gzip > $destdir/$newfile_base_gz") == 0
-	  or die "repacking from bzip2 to gzip failed\n";
+	my (undef, $fname) = tempfile(UNLINK => 1);
+	fork_and_exec(exec => ['bunzip2', '-c', "$destdir/$newfile_base"],
+		      to_file => $fname,
+		      wait_child => 1);
+	fork_and_exec(exec => ['gzip', '-n', '-9'],
+		      from_file => $fname,
+		      to_file => "$destdir/$newfile_base_gz",
+		      wait_child => 1);
+	unlink "$destdir/$newfile_base";
+	$newfile_base = $newfile_base_gz;
+    }
+
+    if ($repack and $newfile_base =~ /^(.*)\.(tar\.lzma|tlz(?:ma?)?)$/) {
+	print "-- Repacking from lzma to gzip\n" if $verbose;
+	my $newfile_base_gz = "$1.tar.gz";
+	my (undef, $fname) = tempfile(UNLINK => 1);
+	fork_and_exec(exec => ['xz', '-F', 'lzma', '-cd', "$destdir/$newfile_base"],
+		      to_file => $fname,
+		      wait_child => 1);
+	fork_and_exec(exec => ['gzip', '-n', '-9'],
+		      from_file => $fname,
+		      to_file => "$destdir/$newfile_base_gz",
+		      wait_child => 1);
 	unlink "$destdir/$newfile_base";
 	$newfile_base = $newfile_base_gz;
     }
@@ -1238,125 +1380,117 @@ EOF
 	  or die("unzip binary not found. You need to install the package unzip to be able to repack .zip upstream archives.\n");
 
 	my $newfile_base_gz = "$1.tar.gz";
-	my $tempdir = tempdir ( "uscanXXXX", TMPDIR => 1, CLEANUP => 1 );
-	system("unzip -q -d $tempdir $destdir/$newfile_base; tar -C $tempdir -czf $destdir/$newfile_base_gz .") == 0 
-	  or die("Repacking from zip to tar.gz failed\n");
+	my $tempdir = tempdir ("uscanXXXX", TMPDIR => 1, CLEANUP => 1);
+	# Parent of the target directory should be under our control
+	$tempdir .= '/repack';
+	mkdir $tempdir or die("Unable to mkdir($tempdir): $!\n");
+	my $absdestdir = abs_path($destdir);
+	system('unzip', '-q', '-a', '-d', $tempdir, "$destdir/$newfile_base") == 0
+	    or die("Repacking from zip to tar.gz failed (could not unzip)\n");
+	my $cwd = cwd();
+	chdir($tempdir) or die("Unable to chdir($tempdir): $!\n");
+	eval {
+	    fork_and_exec(exec => ['tar', '--owner=root', '--group=root', '--mode=a+rX', '-czf', "$absdestdir/$newfile_base_gz", glob('* .[!.]*')],
+			  env => { GZIP => '-n -9' },
+			  wait_child => 1);
+	};
+	if ($@) {
+	    die("Repacking from zip to tar.gz failed (could not create tarball)\n");
+	}
+	chdir($cwd);
 	unlink "$destdir/$newfile_base";
 	$newfile_base = $newfile_base_gz;
     }
 
-    if ($newfile_base =~ /\.(tar\.gz|tgz|tar\.bz2|tbz2?)$/) {
-	my $filetype = `file $destdir/$newfile_base`;
-	$filetype =~ s%^\.\./\Q$newfile_base\E: %%;
-	unless ($filetype =~ /compressed data/) {
+    if ($newfile_base =~ /\.(tar\.gz|tgz
+			     |tar\.bz2|tbz2?
+			     |tar.lzma|tlz(?:ma?)?)$/x) {
+	my $filetype;
+	eval {
+	    fork_and_exec(exec => ['file', '-b', '-k', "$destdir/$newfile_base"],
+			  to_string => \$filetype,
+			  wait_child => 1);
+	};
+	unless (defined $filetype && $filetype =~ /compressed data/) {
 	    warn "$progname warning: $destdir/$newfile_base does not appear to be a compressed file;\nthe file command says: $filetype\nNot processing this file any further!\n";
 	    return 1;
 	}
     }
 
-    if ($newfile_base =~ /\.(tar\.gz|tgz)$/) {
-	if ($symlink eq 'symlink') {
-	    symlink $newfile_base, "$destdir/${pkg}_${newversion}.orig.tar.gz";
-	} elsif ($symlink eq 'rename') {
-	    move "$destdir/$newfile_base", "$destdir/${pkg}_${newversion}.orig.tar.gz";
-	}
-    }
-    elsif ($newfile_base =~ /\.(tar\.bz2|tbz2?)$/) {
-	if ($symlink eq 'symlink') {
-	    symlink $newfile_base, "$destdir/${pkg}_${newversion}.orig.tar.bz2";
-	} elsif ($symlink eq 'rename') {
-	    move "$destdir/$newfile_base", "$destdir/${pkg}_${newversion}.orig.tar.bz2";
-	}
-    }
+    my @renames = (
+	[qr/\.(tar\.gz|tgz)$/, 'gz'],
+	[qr/\.(tar\.bz2|tbz2?)$/, 'bz2'],
+	[qr/\.tar\.lzma|tlz(?:ma?)?$/, 'lzma'],
+    );
 
-    if ($verbose) {
-	print "-- Successfully downloaded updated package $newfile_base\n";
-	if ($newfile_base =~ /\.(tar\.gz|tgz)$/) {
+    my ($renamed_base);
+    foreach my $pair (@renames) {
+	if ($newfile_base !~ $pair->[0]) {
+	    next;
+	}
+
+	my ($pattern, $suffix) = @{$pair};
+	$renamed_base = "${pkg}_${newversion}.orig.tar.$suffix";
+	if ($symlink eq 'symlink') {
+	    symlink $newfile_base, "$destdir/$renamed_base";
+	} elsif ($symlink eq 'rename') {
+	    move "$destdir/$newfile_base", "$destdir/$renamed_base";
+	}
+	if ($verbose) {
+	    print "-- Successfully downloaded updated package $newfile_base\n";
 	    if ($symlink eq 'symlink') {
-		print "    and symlinked ${pkg}_${newversion}.orig.tar.gz to it\n";
+		print "    and symlinked $renamed_base to it\n";
 	    } elsif ($symlink eq 'rename') {
-		print "    and renamed it as ${pkg}_${newversion}.orig.tar.gz\n";
+		print "    and renamed it as $renamed_base\n";
 	    }
-	}
-	elsif ($newfile_base =~ /\.(tar\.bz2|tbz2?)$/) {
+	} elsif ($dehs) {
+	    my $msg = "Successfully downloaded updated package $newfile_base";
+	    $dehs_tags{'target'} = "$renamed_base";
 	    if ($symlink eq 'symlink') {
-		print "    and symlinked ${pkg}_${newversion}.orig.tar.bz2 to it\n";
+		$msg .= " and symlinked $renamed_base to it";
 	    } elsif ($symlink eq 'rename') {
-		print "    and renamed it as ${pkg}_${newversion}.orig.tar.bz2\n";
+		$msg .= " and renamed it as $renamed_base";
+	    } else {
+		$dehs_tags{'target'} = $newfile_base;
 	    }
-	}
-    } elsif ($dehs) {
-	my $msg = "Successfully downloaded updated package $newfile_base";
-	if ($newfile_base =~ /\.(tar\.gz|tgz)$/) {
+	    dehs_msg($msg);
+	} else {
+	    print "$pkg: Successfully downloaded updated package $newfile_base\n";
 	    if ($symlink eq 'symlink') {
-		$msg .= " and symlinked ${pkg}_${newversion}.orig.tar.gz to it";
+		print "    and symlinked $renamed_base to it\n";
 	    } elsif ($symlink eq 'rename') {
-		$msg .= " and renamed it as ${pkg}_${newversion}.orig.tar.gz";
-	    }
-	}
-	elsif ($newfile_base =~ /\.(tar\.bz2|tbz2?)$/) {
-	    if ($symlink eq 'symlink') {
-		$msg .= " and symlinked ${pkg}_${newversion}.orig.tar.bz2 to it";
-	    } elsif ($symlink eq 'rename') {
-		$msg .= " and renamed it as ${pkg}_${newversion}.orig.tar.bz2";
-	    }
-	}
-	dehs_msg($msg);
-    } else {
-	print "$pkg: Successfully downloaded updated package $newfile_base\n";
-	if ($newfile_base =~ /\.(tar\.gz|tgz)$/) {
-	    if ($symlink eq 'symlink') {
-		print "    and symlinked ${pkg}_${newversion}.orig.tar.gz to it\n";
-	    } elsif ($symlink eq 'rename') {
-		print "    and renamed it as ${pkg}_${newversion}.orig.tar.gz\n";
-	    }
-	}
-	elsif ($newfile_base =~ /\.(tar\.bz2|tbz2?)$/) {
-	    if ($symlink eq 'symlink') {
-		print "    and symlinked ${pkg}_${newversion}.orig.tar.bz2 to it\n";
-	    } elsif ($symlink eq 'rename') {
-		print "    and renamed it as ${pkg}_${newversion}.orig.tar.bz2\n";
+		print "    and renamed it as $renamed_base\n";
 	    }
 	}
+	last;
     }
 
     # Do whatever the user wishes to do
     if ($action) {
 	my $usefile = "$destdir/$newfile_base";
 	my @cmd = shellwords($action);
-	if ($symlink =~ /^(symlink|rename)$/
-	    and $newfile_base =~ /\.(tar\.gz|tgz)$/) {
-	    $usefile = "$destdir/${pkg}_${newversion}.orig.tar.gz";
-	}
-	elsif ($symlink =~ /^(symlink|rename)$/
-	    and $newfile_base =~ /\.(tar\.bz2|tbz2)$/) {
-	    $usefile = "$destdir/${pkg}_${newversion}.orig.tar.bz2";
+	if ($symlink =~ /^(symlink|rename)$/ && $renamed_base) {
+	    $usefile = "$destdir/$renamed_base";
 	}
 
 	# Any symlink requests are already handled by uscan
 	if ($action =~ /^uupdate(\s|$)/) {
-	    $action =~ s/^uupdate/uupdate --no-symlink/;
 	    push @cmd, "--no-symlink";
 	}
 
 	if ($watch_version > 1) {
-	    print "-- Executing user specified script\n     $action --upstream-version $newversion $newfile_base" if $verbose;
-	    if ($dehs) {
-		my $msg = "Executing user specified script: $action --upstream-version $newversion $newfile_base; output:\n";
-		$msg .= `$action --upstream-version $newversion $usefile 2>&1`;
-		dehs_msg($msg);
-	    } else {
-		system(@cmd, "--upstream-version", "$newversion", "$usefile");
-	    }
+	    push @cmd, ("--upstream-version", "$newversion", "$usefile");
 	} else {
-	    print "-- Executing user specified script $action $newfile_base $newversion" if $verbose;
-	    if ($dehs) {
-		my $msg = "Executing user specified script: $action $newfile_base $newversion; output:\n";
-		$msg .= `$action $usefile $newversion 2>&1`;
-		dehs_msg($msg);
-	    } else {
-		system(@cmd, "$usefile", "$newversion");
-	    }
+	    push @cmd, ("$usefile", "$newversion");
+	}
+	my $actioncmd = join(" ", @cmd);
+	print "-- Executing user specified script\n     $actioncmd\n" if $verbose;
+	if ($dehs) {
+	    my $msg = "Executing user specified script: $actioncmd; output:\n";
+	    $msg .= `$actioncmd 2>&1`;
+	    dehs_msg($msg);
+	} else {
+	    system(@cmd);
 	}
     }
 
@@ -1484,9 +1618,9 @@ sub newest_dir ($$$$$) {
 
 	# We separate out HTMLised listings from standard listings, so
 	# that we can target our search correctly
-	if (/<\s*a\s+[^>]*href/i) {
+	if ($content =~ /<\s*a\s+[^>]*href/i) {
 	    while ($content =~ 
-		m/(?:<\s*a\s+[^>]*href\s*=\s*\")($pattern)\"/gi) {
+		m/(?:<\s*a\s+[^>]*href\s*=\s*\")((?-i)$pattern)\"/gi) {
 		my $dir = $1;
 		my $mangled_version = join(".", $dir =~ m/^$pattern$/);
 		push @dirs, [$mangled_version, $dir];
@@ -1494,7 +1628,7 @@ sub newest_dir ($$$$$) {
 	} else {
 	    # they all look like:
 	    # info info ... info filename [ -> linkname]
-	    while ($content =~ m/($pattern)(\s+->\s+\S+)?$/mgi) {
+	    while ($content =~ m/($pattern)(\s+->\s+\S+)?$/mg) {
 		my $dir = $1;
 		my $mangled_version = join(".", $dir =~ m/^$pattern$/);
 		push @dirs, [$mangled_version, $dir];
@@ -1509,7 +1643,7 @@ sub newest_dir ($$$$$) {
 	    my ($newversion, $newdir) = @{$dirs[0]};
 	    return $newdir;
 	} else {
-	    warn "$progname warning: In $watchfile no matching dirs for pattern\n  $site$base$pattern\n";
+	    warn "$progname warning: In $watchfile no matching dirs for pattern\n  $base$pattern\n";
 	    return '';
 	}
     }
@@ -1841,7 +1975,7 @@ sub safe_replace($$) {
 	    # Actually do the replacement
 	    substr $$in, $first, $last - $first, $replacement;
 	    # Update position
-	    $pos += length($replacement)- ($last - $first);
+	    $pos += length($replacement) - ($last - $first);
 
 	    if ($global) {
 		$replacement = $orig_replacement;
